# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2017-2018  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
# Copyright (C) 2019  Tomasz Magdziarz <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
#

"""
Module implements AutoBarber generation of spheres.
"""

from aquaduct import logger
from scipy.spatial.distance import cdist
import numpy as np

from aquaduct.utils import clui
from aquaduct.utils.helpers import listify
from aquaduct.utils.helpers import lind, SmartRange
from aquaduct.traj.sandwich import ReaderAccess, Reader
from aquaduct.geom import Sphere, do_cut_thyself
from aquaduct.utils.multip import optimal_threads
from multiprocessing import Pool
from functools import partial
from itertools import chain, izip, imap
from aquaduct.apps.data import CRIC

__mail__ = 'info@aquaduct.pl'


@listify
def spaths2spheres(spaths, minmax=None, selection=None, tovdw=None, forceempty=None):
    mincut, mincut_val, maxcut, maxcut_val, mincut_level, maxcut_level = minmax
    for sp in spaths:

        traj_reader = Reader.get_reader_by_id(sp.id.id).open()
        barber = traj_reader.parse_selection(selection)

        vdwradius = 0

        centers = []
        frames = []
        # TODO: This is inconsistent with inlets types. Below is equivalent to surface only inlets. Rework this and make it coherent to each other.
        # Assume it is already coherent.
        if sp.has_in:
            centers.append(sp.coords_first_in)
            frames.append(sp.paths_first_in)
        if sp.has_out:
            centers.append(sp.coords_last_out)
            frames.append(sp.paths_last_out)
        for center, frame in zip(centers, frames):
            make_sphere = True
            if make_sphere:
                traj_reader.set_frame(frame)
                distances = cdist(np.matrix(center), np.matrix(list(barber.coords())), metric='euclidean').flatten()
                if tovdw:
                    vdwradius = list(barber.ix(np.argmin(distances)).vdw())[0]
                    logger.debug('VdW correction %0.2f', vdwradius)
                radius = min(distances) - vdwradius
                if radius <= 0:
                    logger.debug('VdW correction resulted in <= 0 radius.')
                    make_sphere = False
                if mincut and radius < mincut_val:
                    if not mincut_level:
                        logger.debug('Sphere radius %0.2f is less then mincut %0.2f', radius, mincut_val)
                        make_sphere = False
                    else:
                        logger.debug('Sphere radius %0.2f leveled to mincut %0.2f', radius, mincut_val)
                        radius = mincut_val
                if maxcut and radius > maxcut_val:
                    if not maxcut_level:
                        logger.debug('Sphere radius %0.2f is greater then maxcut %0.2f', radius, maxcut_val)
                        make_sphere = False
                    else:
                        logger.debug('Sphere radius %0.2f leveled to maxcut %0.2f', radius, maxcut_val)
                        radius = maxcut_val
            if make_sphere:
                logger.debug('Added sphere of radius %0.2f' % radius)
                yield (center.flatten(), radius)
            elif forceempty:
                logger.debug('Added sphere of radius 0')
                yield (center.flatten(), 0)
    yield CRIC


@listify
def inlets2spheres(inlets, minmax=None, selection=None, tovdw=None, forceempty=None):
    mincut, mincut_val, maxcut, maxcut_val, mincut_level, maxcut_level = minmax
    for inlet in inlets:

        traj_reader = Reader.get_reader_by_id(inlet.reference.id).open()
        barber = traj_reader.parse_selection(selection)

        vdwradius = 0

        center = inlet.coords
        frame = inlet.frame

        make_sphere = True
        if make_sphere:
            traj_reader.set_frame(frame)
            distances = cdist(np.matrix(center), np.matrix(list(barber.coords())), metric='euclidean').flatten()
            if tovdw:
                vdwradius = list(barber.ix(np.argmin(distances)).vdw())[0]
                logger.debug('VdW correction %0.2f', vdwradius)
            radius = min(distances) - vdwradius
            if radius <= 0:
                logger.debug('VdW correction resulted in <= 0 radius.')
                make_sphere = False
            if mincut and radius < mincut_val:
                if not mincut_level:
                    logger.debug('Sphere radius %0.2f is less then mincut %0.2f', radius, mincut_val)
                    make_sphere = False
                else:
                    logger.debug('Sphere radius %0.2f leveled to mincut %0.2f', radius, mincut_val)
                    radius = mincut_val
            if maxcut and radius > maxcut_val:
                if not maxcut_level:
                    logger.debug('Sphere radius %0.2f is greater then maxcut %0.2f', radius, maxcut_val)
                    make_sphere = False
                else:
                    logger.debug('Sphere radius %0.2f leveled to maxcut %0.2f', radius, maxcut_val)
                    radius = maxcut_val
        if make_sphere:
            logger.debug('Added sphere of radius %0.2f' % radius)
            yield (center.flatten(), radius)
        elif forceempty:
            logger.debug('Added sphere of radius 0')
            yield (center.flatten(), 0)
    yield CRIC


class WhereToCut(ReaderAccess):
    '''
    Class implements method for creating (optimal) set of AutoBarber spheres for a collection of spaths;
    access to trajectory is also required to read VdW radii.
    '''

    # creates collection of Spheres
    def __init__(self,
                 spaths=None,
                 inlets=None,
                 expected_nr_of_spaths=None,
                 selection=None,
                 mincut=None,
                 mincut_level=False,
                 maxcut=None,
                 maxcut_level=False,
                 tovdw=False,
                 forceempty=False):
        '''
        :param list spaths: List of :class:`aquaduct.traj.paths.SinglePath` objects.
        :param int expected_nr_of_spaths: Number of spaths passed. Requilred when length of spaths cannod be calculated, eg when it is a generator.
        :param str selection: Selection string of molecular object used for spheres generation.
        :param float mincut: Value of *mincut* parameter.
        :param float maxcut: Value of *maxcut* parameter.
        :param bool mincut_level: Flag of *mincut_level*.
        :param bool maxcut_level: Flag of *maxcut_level*.
        :param bool tovdw: Flag of to VdW radii correction parameter.
        :param bool forceemtpy: If set *True* spheres of radius 0 are returned if no other sphere can be generated.

        '''
        # force empty means that empty spheres are also returned with radius 0
        self.forceempty = forceempty
        self.selection = selection
        self.mincut = mincut
        self.maxcut = maxcut
        self.mincut_level = mincut_level
        self.maxcut_level = maxcut_level
        self.tovdw = tovdw

        self.expected_nr_of_spaths = expected_nr_of_spaths

        self.spheres = []

        if spaths is not None:
            self.add_spheres_from_spaths(spaths)
        if inlets is not None:
            self.add_spheres_from_inlets(inlets)

    def check_minmaxcuts(self):
        mincut = False
        mincut_val = None
        if self.mincut is not None:
            mincut = True
            mincut_val = float(self.mincut)
            logger.debug('mincut set to %0.2f', mincut_val)
        maxcut = False
        maxcut_val = None
        if self.maxcut is not None:
            maxcut = True
            maxcut_val = float(self.maxcut)
            logger.debug('maxcut set to %0.2f', maxcut_val)
        if mincut and maxcut:
            if maxcut_val < mincut_val:
                logger.warning(
                    'Values of mincut %0.2f and maxcut %0.2f are mutually exclusive. No spheres will be used in Auto Barber.',
                    mincut_val, maxcut_val)
                if self.maxcut_level or self.mincut_level:
                    logger.warning(
                        'Options mincut_level and maxcut_level are set to %s and %s accordingly, some spheres might be retained.')
        return mincut, mincut_val, maxcut, maxcut_val

    def add_spheres_from_spaths(self, spaths):
        clui.message("Auto Barber is looking where to cut:")
        if self.expected_nr_of_spaths:
            pbar = clui.pbar(self.expected_nr_of_spaths)
            n_add = self.expected_nr_of_spaths
        else:
            pbar = clui.pbar(len(spaths))
            n_add = len(spaths)
        minmax = self.check_minmaxcuts() + (self.mincut_level, self.maxcut_level)

        Reader.reset()
        pool = Pool(processes=optimal_threads.threads_count)
        n = max(1, optimal_threads.threads_count)
        chunks = (n if chunk <= (n_add / n - 1) else (n_add % n) for chunk in xrange(n_add / n + np.sign(n_add % n)))

        add_function = partial(spaths2spheres, minmax=minmax, selection=self.selection, tovdw=self.tovdw,
                               forceempty=self.forceempty)
        _spaths = chain(spaths)
        Reader.reset()
        spheres_new = pool.imap(add_function, ([_spaths.next() for cc in xrange(c)] for c in chunks))

        nr = 0
        for spheres in spheres_new:
            for center_radius_or_cric in spheres:
                if isinstance(center_radius_or_cric, tuple):
                    center, radius = center_radius_or_cric
                    self.spheres.append(Sphere(center=center, radius=radius, nr=self.get_current_nr()))
                else:
                    CRIC.update_cric(center_radius_or_cric)
            nr += n
            if nr < n_add:
                pbar.update(nr)
            else:
                pbar.update(n_add)
        pool.close()
        pool.join()
        pbar.finish()

    def add_spheres_from_inlets(self, inlets):
        clui.message("Auto Barber is looking where to cut:")
        if self.expected_nr_of_spaths:
            pbar = clui.pbar(self.expected_nr_of_spaths)
            n_add = self.expected_nr_of_spaths
        else:
            pbar = clui.pbar(len(inlets.inlets_list))
            n_add = len(inlets.inlets_list)
        minmax = self.check_minmaxcuts() + (self.mincut_level, self.maxcut_level)

        Reader.reset()
        pool = Pool(processes=optimal_threads.threads_count)
        # pool = Pool(processes=1)
        n = max(1, optimal_threads.threads_count)
        chunks = (n if chunk <= (n_add / n - 1) else (n_add % n) for chunk in xrange(n_add / n + np.sign(n_add % n)))

        add_function = partial(inlets2spheres, minmax=minmax, selection=self.selection, tovdw=self.tovdw,
                               forceempty=self.forceempty)
        _inlets = chain(inlets)
        spheres_new = pool.imap(add_function, [[_inlets.next() for cc in xrange(c)] for c in chunks])
        # spheres_new = imap(add_function, ([_inlets.next() for cc in xrange(c)] for c in chunks))

        nr = 0
        for spheres in spheres_new:
            for center_radius_or_cric in spheres:
                if isinstance(center_radius_or_cric, tuple):
                    center, radius = center_radius_or_cric
                    self.spheres.append(Sphere(center=center, radius=radius, nr=self.get_current_nr()))
                else:
                    CRIC.update_cric(center_radius_or_cric)
            nr += n
            if nr < n_add:
                pbar.update(nr)
            else:
                pbar.update(n_add)
        pool.close()
        pool.join()
        pbar.finish()

    def get_current_nr(self):
        if len(self.spheres):
            # nr = max((sphe.nr for sphe in self.spheres))
            nr = self.spheres[-1].nr
            nr += 1
        else:
            nr = 0
        assert nr >= len(self.spheres), "Inconsistent number of spheres."
        return nr

    def inlet2sphere(self, inlet):
        traj_reader = Reader.get_reader_by_id(inlet.reference.id).open()
        mincut, mincut_val, maxcut, maxcut_val = self.check_minmaxcuts()
        barber = traj_reader.parse_selection(self.selection)
        vdwradius = 0

        center = inlet.coords
        frame = inlet.frame

        make_sphere = True
        if make_sphere:
            traj_reader.set_frame(frame)
            distances = cdist(np.matrix(center), np.matrix(list(barber.coords())), metric='euclidean').flatten()
            if self.tovdw:
                vdwradius = list(barber.ix(np.argmin(distances)).vdw())[0]
                logger.debug('VdW correction %0.2f', vdwradius)
            radius = min(distances) - vdwradius
            if radius <= 0:
                logger.debug('VdW correction resulted in <= 0 radius.')
                make_sphere = False
            if mincut and radius < mincut_val:
                if not self.mincut_level:
                    logger.debug('Sphere radius %0.2f is less then mincut %0.2f', radius, mincut_val)
                    make_sphere = False
                else:
                    logger.debug('Sphere radius %0.2f leveled to mincut %0.2f', radius, mincut_val)
                    radius = mincut_val
            if maxcut and radius > maxcut_val:
                if not self.maxcut_level:
                    logger.debug('Sphere radius %0.2f is greater then maxcut %0.2f', radius, maxcut_val)
                    make_sphere = False
                else:
                    logger.debug('Sphere radius %0.2f leveled to maxcut %0.2f', radius, maxcut_val)
                    radius = maxcut_val
        if make_sphere:
            logger.debug('Added sphere of radius %0.2f' % radius)
            return Sphere(center.flatten(), radius, self.get_current_nr())
        elif self.forceempty:
            logger.debug('Added sphere of radius 0')
            return Sphere(center.flatten(), 0, self.get_current_nr())

    def spath2spheres(self, sp):

        traj_reader = self.get_reader_by_id(sp.id.id)
        mincut, mincut_val, maxcut, maxcut_val = self.check_minmaxcuts()
        barber = traj_reader.parse_selection(self.selection)
        vdwradius = 0

        centers = []
        frames = []
        # TODO: This is inconsistent with inlets types. Below is equivalent to surface only inlets. Rework this and make it coherent to each other.
        # Assume it is already coherent.
        if sp.has_in:
            centers.append(sp.coords_first_in)
            frames.append(sp.paths_first_in)
        if sp.has_out:
            centers.append(sp.coords_last_out)
            frames.append(sp.paths_last_out)
        for center, frame in zip(centers, frames):
            make_sphere = True
            if make_sphere:
                traj_reader.set_frame(frame)
                distances = cdist(np.matrix(center), np.matrix(list(barber.coords())), metric='euclidean').flatten()
                if self.tovdw:
                    vdwradius = list(barber.ix(np.argmin(distances)).vdw())[0]
                    logger.debug('VdW correction %0.2f', vdwradius)
                radius = min(distances) - vdwradius
                if radius <= 0:
                    logger.debug('VdW correction resulted in <= 0 radius.')
                    make_sphere = False
                if mincut and radius < mincut_val:
                    if not self.mincut_level:
                        logger.debug('Sphere radius %0.2f is less then mincut %0.2f', radius, mincut_val)
                        make_sphere = False
                    else:
                        logger.debug('Sphere radius %0.2f leveled to mincut %0.2f', radius, mincut_val)
                        radius = mincut_val
                if maxcut and radius > maxcut_val:
                    if not self.maxcut_level:
                        logger.debug('Sphere radius %0.2f is greater then maxcut %0.2f', radius, maxcut_val)
                        make_sphere = False
                    else:
                        logger.debug('Sphere radius %0.2f leveled to maxcut %0.2f', radius, maxcut_val)
                        radius = maxcut_val
            if make_sphere:
                logger.debug('Added sphere of radius %0.2f' % radius)
                yield Sphere(center.flatten(), radius, self.get_current_nr())
            elif self.forceempty:
                logger.debug('Added sphere of radius 0')
                yield Sphere(center.flatten(), 0, self.get_current_nr())

    def cut_thyself(self):
        self.spheres = do_cut_thyself(self.spheres, progress=True)[0]

    def is_overlaping_with_cloud(self, sphere):
        spheres_coords = np.array([sphe.center for sphe in self.spheres])
        spheres_radii = np.array([sphe.radius for sphe in self.spheres])
        center, radius, nr = sphere
        distances = cdist(np.matrix(center), spheres_coords, metric='euclidean').flatten()
        distances = distances - spheres_radii - radius
        return (distances <= 0).any()

    def cloud_groups(self, progress=False):
        # no redundant spheres
        noredundant_spheres, redundant_spheres = do_cut_thyself(self.spheres, progress=progress)
        if progress:
            clui.message("Barber, clouds clusters:")
            pbar = clui.pbar(len(self.spheres))

        noredundant_spheres_coords = np.array([sphe.center for sphe in noredundant_spheres])
        noredundant_spheres_radii = np.array([sphe.radius for sphe in noredundant_spheres])
        clouds = {}
        for nrs in noredundant_spheres:
            # calculate distance
            center, radius, nr = nrs
            distances = cdist(np.matrix(center), noredundant_spheres_coords, metric='euclidean').flatten()
            distances = distances - noredundant_spheres_radii - radius
            current_cloud = set(np.argwhere(distances <= 0).flatten().tolist())
            del distances
            # check if cci overlaps with any of already found clouds
            cloud_id_intersections = []
            for cloud_id, cloud in clouds.iteritems():
                if current_cloud.intersection(cloud):
                    # current cloud intersects with cloud
                    cloud_id_intersections.append(cloud_id)
            # does it intersects?
            if cloud_id_intersections:
                for cii in cloud_id_intersections:
                    current_cloud = current_cloud.union(clouds.pop(cii))
                    # current id?
            current_id = clouds.keys()
            if current_id:
                for cid in range(max(current_id) + 2):
                    if cid not in current_id:
                        current_id = cid
                        break
            else:
                current_id = 0
            clouds.update({current_id: current_cloud})
            if progress:
                pbar.next()

        # chnage nrs id to global ids; add redundant spheres
        nrs_gids = [nrs.nr for nrs in noredundant_spheres]
        for cloud_id, cloud in clouds.iteritems():

            cloud = sorted(list(cloud))

            if len(redundant_spheres):
                cloud_coords = noredundant_spheres_coords[cloud]
                cloud_radii = noredundant_spheres_radii[cloud]

            cloud = sorted([nrs_gids[c] for c in cloud])
            # clouds.update({cloud_id:cloud})

            for rs_id in range(len(redundant_spheres))[::-1]:
                center, radius, nr = redundant_spheres[rs_id]
                distances = cdist(np.matrix(center), cloud_coords, metric='euclidean').flatten()
                distances = distances - cloud_radii - radius
                current_cloud = np.argwhere(distances <= 0).flatten().tolist()
                if current_cloud:
                    cloud.append(nr)
                    redundant_spheres.pop(rs_id)
                    if progress:
                        pbar.next()

            clouds.update({cloud_id: cloud})

        if progress:
            pbar.finish()

        return clouds


def barber_with_spheres(coords, spheres):
    '''
    param numpy.ndarray coords: Path's coordinates subjected to barber procedure.
    param spheres: Spheres used to cut input coordinates.
    rtype: :class:`numpy.ndarray`.
    return: List of indices to be kept.
    '''
    tokeep = np.ones(len(coords), dtype=np.bool)
    for sp in spheres:
        tokeep = np.logical_and(tokeep, cdist(coords, np.array([sp.center]), metric='euclidean').flatten() > sp.radius)
    return np.argwhere(tokeep).flatten().tolist()


def barber_with_spheres_big_matrix(coords, spheres):
    # calculate big distance matrix
    # distances = cdist(self.coords, [s.center for s in spheres],metric='euclidean')
    # compare with radii
    # tokeep = distances > np.matrix([[s.radius for s in spheres]])
    if len(spheres):
        return np.argwhere((cdist(np.array(list(coords)), [s.center for s in spheres], metric='euclidean') > np.matrix(
            [[s.radius for s in spheres]])).all(1).A1).flatten().tolist()


@listify
def barber_paths(paths, spheres=None, only_for_names=None):
    # cut paths with barber
    for path in (pat for pat in paths if pat.name in only_for_names):
        tokeep = barber_with_spheres(path.coords, spheres)
        path.update_types_frames(SmartRange(lind(path.types, tokeep)), SmartRange(lind(path.frames, tokeep)))
        yield path
    yield CRIC



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2018-2019  Tomasz Magdziarz <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Module performs HDR 2D calculations only with Gaussian Kerneld Density Estimator
as impelemented in :mod:`scipy.stats`.
'''
from itertools import imap
import numpy as np
from scipy.stats import gaussian_kde
from aquaduct.geom.pca import PCA, Polarize
from aquaduct.geom.traces import square_area


class HDR(object):
    def __init__(self, X, points=10, expand_by=1., center_of_system=None):
        # expand_by in A is added to grid boudaries to expand it a bit

        self.center_of_system = center_of_system

        # calcualte PCA
        self.pca = PCA(preprocess=Polarize(center=center_of_system, rvar=0.01))
        self.pca.build(X)

        # get kernel for PC1 and PC2
        self.kernel = gaussian_kde(self.pca.T[:, [0, 1]].T)

        # find x and y ranges
        # xyr = np.array([self.pca.T[:,:2].max(0),self.pca.T[:,:2].min(0)]).mean(0) # center of range
        xyr = (self.pca.T[:, :2].max(0) - self.pca.T[:, :2].min(0) + expand_by) / 2.  # 1/2 expanded range
        self.xyr = np.array([np.array([self.pca.T[:, :2].max(0), self.pca.T[:, :2].min(0)]).mean(0) - xyr,
                             np.array([self.pca.T[:, :2].max(0), self.pca.T[:, :2].min(0)]).mean(0) + xyr])

        # desired number of points in the grid for each of two dimensions per A

        # total number of cells is points^2
        xy = self.xyr[1] - self.xyr[0]
        points = xy.mean() * points

        b = (((points ** 2) * xy[1]) / (xy[1] * (xy[0] ** 2))) ** 0.5
        a = xy[0] / xy[1] * b

        self.points = (xy * np.array([a, b]))

        # calcualte spatial span
        self.X, self.Y = np.mgrid[self.pca.T.min(0)[0]:self.pca.T.max(0)[0]:(self.points[0] * 1j),
                         self.pca.T.min(0)[1]:self.pca.T.max(0)[1]:(self.points[1] * 1j)]
        self.positions = np.vstack([self.X.ravel(), self.Y.ravel()])

        self.values = self.kernel(self.positions)

        self.Z = np.reshape(self.values, self.X.shape)

    @property
    def cell_area(self):
        return self.cell_dimensions.prod()

    @property
    def cell_dimensions(self):
        return (self.pca.T[:, :2].max(0) - self.pca.T[:, :2].min(0)) / self.points

    def _fraction2indices(self, fraction=0.9):
        # indices are returned in the order of values, i.e. after ravel.
        return self.values > self.Z.max() * (1 - fraction)

    def area_naive(self, fraction=0.9):
        '''
        This is naive implementation. It calculates are in the space of prepocessed variables.
        Transformation to normal Cartesian space may not be trivial as details mey depends on the
        actual location  of points.

        :param float fraction: Fraction of cluster for which are should be estimated.
        :rtype: float
        :return: Area of cluster calculated in the preprocessed space.
        '''
        return self.cell_area * (self._fraction2indices(fraction=fraction)).sum()

    def _square_me(self, position):
        return self.pca.undo(position - self.cell_dimensions / 2 * np.array([[1, 1], [1, -1], [-1, -1], [-1, 1]]),
                             pc=[0, 1])

    def area(self, fraction=0.9):
        '''
        :param float fraction: Fraction of cluster for which are should be estimated.
        :rtype: float
        :return: Area of cluster calculated in the Cartesian space.
        '''
        i = self._fraction2indices(fraction=fraction)
        return sum((square_area(*tuple(sq)) for sq in imap(self._square_me, self.positions.T[i])))



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2018  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
# Copyright (C) 2019  Tomasz Magdziarz <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import cPickle as pickle
import os
import tarfile

import aquaduct.visual.pymol_cgo as cgo

from aquaduct.geom import traces
from aquaduct.traj.paths import PathTypesCodes
from aquaduct.utils.helpers import list_blocks_to_slices
from aquaduct.visual.helpers import color_codes, cc
from aquaduct.utils.helpers import create_tmpfile, listify


class BasicPymolCGO(object):
    cgo_entity_begin = []
    cgo_entity_end = []

    def __init__(self):
        self.cgo_entity = None
        self.previous = None
        self.clean()

    def clean(self, empty=False):
        self.previous = None
        self.cgo_entity = []
        if not empty:
            self.cgo_entity.extend(self.cgo_entity_begin)

    def new(self):
        self.previous = None

    def get(self):
        return self.cgo_entity + self.cgo_entity_end

    @staticmethod
    def make_color_triple(color_definition):
        return tuple(color_definition)


class BasicPymolCGOLines(BasicPymolCGO):
    cgo_entity_begin = [cgo.BEGIN, cgo.LINES]
    cgo_entity_end = [cgo.END]

    def add(self, coords=None, color=None):

        if color is not None:
            self.cgo_entity.append(cgo.COLOR)
            # self.cgo_entity.extend(map(float, color))
            self.cgo_entity.append(self.make_color_triple(map(float, color)))

        if coords is not None:
            for nr, coord in enumerate(coords):
                if self.previous is not None:
                    self.cgo_entity.append(cgo.VERTEX)
                    self.cgo_entity.extend(map(float, self.previous))

                    self.cgo_entity.append(cgo.VERTEX)
                    self.cgo_entity.extend(map(float, coord))
                self.previous = coord


class BasicPymolCGOSpheres(BasicPymolCGO):
    cgo_entity_begin = []
    cgo_entity_end = []

    def add(self, coords=None, radius=None, color=None):
        # color to colors...
        if color is not None:
            color = np.matrix(color).A
        if radius is not None:
            radius = np.matrix(radius).A1

        if coords is not None:
            for nr, coord in enumerate(coords):
                if color is not None:
                    if len(color) > 1:
                        c = color[nr]
                    else:
                        c = color[0]
                    self.cgo_entity.append(cgo.COLOR)
                    self.cgo_entity.append(self.make_color_triple(map(float, c)))
                    # self.cgo_entity.extend(map(float, c))
                self.cgo_entity.append(cgo.SPHERE)
                self.cgo_entity.extend(map(float, coord))
                if radius is not None:
                    if len(radius) > 1:
                        r = radius[nr]
                    else:
                        r = radius
                else:
                    r = 1.
                self.cgo_entity.append(float(r))


# [cgo.CONE] + xyz3 + xyz2 + [hradius, 0.0] + color2 + color2

class BasicPymolCGOPointers(BasicPymolCGO):
    cgo_entity_begin = []
    cgo_entity_end = []

    def add_cone(self, coords1=None, coords2=None, radius1=None, radius2=None, color1=None, color2=None):
        # color to colors... ???
        if coords1 is not None and coords2 is not None:
            self.cgo_entity.append(cgo.CONE)
            self.cgo_entity.extend(map(float, coords1))
            self.cgo_entity.extend(map(float, coords2))
            self.cgo_entity.append(float(radius1))
            self.cgo_entity.append(float(radius2))
            self.cgo_entity.append(self.make_color_triple(map(float, color1)))
            self.cgo_entity.append(self.make_color_triple(map(float, color2)))
            # self.cgo_entity.extend(map(float, color1))
            # self.cgo_entity.extend(map(float, color2))

            self.cgo_entity.extend([cgo.NULL, cgo.POINTS])

    def add_pointer(self, point=None, direction=None, length=None, color=None, reverse=False):
        vec = point - direction
        vec_len = np.sqrt(np.sum(vec ** 2))
        vec = (vec / vec_len) * length
        vec = point + vec
        if reverse:
            self.add_cone(coords1=point, coords2=vec, radius1=length / 3., radius2=0, color1=color, color2=color)
        else:
            self.add_cone(coords1=vec, coords2=point, radius1=length / 3., radius2=0, color1=color, color2=color)


class SimpleTarWriteHelper(object):
    def __init__(self):
        self.tar_fh = None
        self.tmp_file = create_tmpfile()

    def open(self, filename):
        self.tar_fh = tarfile.open(filename, 'w:gz')

    def save_object2tar(self, obj, name):
        with open(self.tmp_file, 'w') as f:
            pickle.dump(obj, f)
        self.save_file2tar(self.tmp_file, name)

    def save_file2tar(self, filename, name):
        self.tar_fh.add(filename, arcname=name)

    def __del__(self):
        if self.tar_fh is not None:
            self.tar_fh.close()
        os.unlink(self.tmp_file)


class ConnectToPymol(object):
    cgo_line_width = 2.
    ct_pymol = 'pymol'
    ct_file = 'file'

    def __init__(self):
        self.connection_type = None  # possible types are pymol and file

        self.script_fh = None
        self.data_fh = SimpleTarWriteHelper()

        self.cmd = None

    @listify
    def decode_color(self, cgo_object):
        for element in cgo_object:
            if isinstance(element, tuple):
                for e in element:
                    yield e
            else:
                yield element

    def init_pymol(self):
        import pymol
        self.cmd = pymol.cmd
        pymol.finish_launching()
        self.cmd.set('cgo_line_width', ConnectToPymol.cgo_line_width)
        self.connection_type = self.ct_pymol

    def init_script(self, filename):
        self.script_fh = open(filename, 'w')
        data_filename = os.path.splitext(os.path.basename(filename))[0] + '.tar.gz'
        self.data_fh.open(data_filename)
        self.connection_type = self.ct_file
        # TODO: Script generated by this object does not work properly if run via PyMOL.
        # init lines, imports etc.
        self.script_fh.write('''import argparse
parser=argparse.ArgumentParser(description="Aqua-Duct visualization script")
parser.add_argument("--save-session",action="store",dest="session",required=False,default=None,help="Pymol session file name.")
parser.add_argument("--discard",action="store",dest="discard",required=False,default='',help="Objects to discard.")
parser.add_argument("--keep",action="store",dest="keep",required=False,default='',help="Objects to keep.")
parser.add_argument("--force-color",action="store",dest="fc",required=False,default='',help="Force specific color.")
parser.add_argument("--fast",action="store_true",dest="fast",required=False,help="Hides all objects while loading.")
args,unknown=parser.parse_known_args()
import sys
if unknown: print >> sys.stderr, "WARNING: Unknown options were used: "+" ".join(unknown)
def _kd_order():
    if args.keep=='' and args.discard!='': return 'd'
    if args.keep!='' and args.discard=='': return 'k'
    if args.keep=='' and args.discard=='': return None
    if sys.argv.index('--keep')<sys.argv.index('--discard'): return 'k'
    return 'd'
kd_order = _kd_order()
def discard(name):
    if len([d for d in args.discard.split() if d in name])>0: return True
    return False
def keep(name):
    if len([k for k in args.keep.split() if k in name])>0: return True
    return False
def proceed(name):
    if kd_order == 'k':
        if not keep(name): return False
        elif discard(name): return False
    elif kd_order == 'd':
        if discard(name):
            if not keep(name): return False
    return True
from pymol import cmd,finish_launching
finish_launching()
print "Loading Aqua-Duct visualization..."
cmd.set("cgo_line_width",%d)
cmd.set("line_smooth","off")
from os import close,unlink
from os.path import splitext,isfile
import tarfile
import cPickle as pickle
from tempfile import mkstemp
fd, pdb_filename = mkstemp(suffix=".pdb")
close(fd)
max_state=0
arch_file="%s"
if not isfile(arch_file):
    import pymol
    if pymol.IS_WINDOWS:
        print "Please open visualization script using 'Open with' context menu and choose PyMol executable."
        print "Alternatively, if you have PyMol installed as Python module, open visulaization script with Python executable."
    while (pymol._ext_gui is None): pymol = reload(pymol)
    while (not hasattr(pymol._ext_gui,'root')): pymol = reload(pymol)
    import tkFileDialog
    arch_file=tkFileDialog.askopenfilename(filetypes=[("AQ Vis Arch","*.tar.gz")],title="Select AQ visualization archive",parent=pymol._ext_gui.root)
data_fh=tarfile.open(arch_file,"r:gz")
def decode_color(cgo_object,fc=None):
    for element in cgo_object:
        if isinstance(element,tuple):
            if fc is None:
                for e in element: yield e
            else:
                for e in fc: yield e
        else:
            yield element
def load_object(filename,name,state):
    if not proceed(name): return
    global max_state
    print "Loading %s" % splitext(filename)[0]
    obj=pickle.load(data_fh.extractfile(filename))
    if name in args.fc.split():
        forced_color=args.fc.split()[args.fc.split().index(name)+1]
        forced_color=cmd.get_color_tuple(forced_color)
        obj=decode_color(obj,fc=forced_color)
    else:
        obj=decode_color(obj)
    cmd.load_cgo(obj,name,state)
    if state<2:
        if args.fast: cmd.disable("all")
        else: cmd.refresh()
    if state>max_state:
        max_state=state
def load_pdb(filename,name,state):
    if not proceed(name): return
    global max_state
    with open(pdb_filename,'w') as fpdb:
        fpdb.write(data_fh.extractfile(filename).read())
    cmd.load(pdb_filename,state=state,object=name)
    if state>max_state:
        max_state=state
''' % (self.cgo_line_width, data_filename, "%s", "% s"))

    def add_cgo_object(self, name, cgo_object, state=None):
        if state is None:
            state = 1
        if self.connection_type == self.ct_pymol:
            self.cmd.load_cgo(self.decode_color(cgo_object), str(name), state)
        elif self.connection_type == self.ct_file:
            obj_name = '%s_%d.dump' % (name, state)
            self.data_fh.save_object2tar(cgo_object, obj_name)

            self.script_fh.write('''load_object("%s","%s",%d)''' % (obj_name, str(name), state))
            self.script_fh.write(os.linesep)
            # self.script_fh.write('''self.cmd.refresh()''')
            # self.script_fh.write(os.linesep)

    def del_cgo_object(self, name, state=None):
        raise NotImplementedError("Deletion of CGO objects is not implemented yet.")

    def load_pdb(self, name, filename, state=None):
        if state is None:
            state = 1
        if self.connection_type == self.ct_pymol:
            self.cmd.load(filename, state=state, object=name)
            self.cmd.show_as('cartoon', name)
            self.cmd.color('silver', name)
        elif self.connection_type == self.ct_file:
            # save pdblile as string
            filename_new = '%s_%d.pdb' % (name, state)
            self.data_fh.save_file2tar(filename, filename_new)
            self.script_fh.write('''load_pdb("%s","%s",%d)''' % (filename_new, name, state))
            self.script_fh.write(os.linesep)
            self.script_fh.write('''if proceed("%s"): cmd.show_as('cartoon','%s')''' % (name, name))
            self.script_fh.write(os.linesep)
            self.script_fh.write('''if proceed("%s"): cmd.color('silver','%s')''' % (name, name))
            self.script_fh.write(os.linesep)

    def orient_on(self, name):
        if self.connection_type == self.ct_pymol:
            self.cmd.orient(name)
        elif self.connection_type == self.ct_file:
            self.script_fh.write('''if proceed("%s"): cmd.orient("%s")''' % (name, name))
            self.script_fh.write(os.linesep)

    def __del__(self):
        if self.connection_type == self.ct_file:
            self.script_fh.write('''data_fh.close()
unlink(pdb_filename)
if args.fast: cmd.enable("all")
print "Aqua-Duct visualization loaded."
if args.session:
    print "Preparing data to save session..."
    for state in range(max_state):
        cmd.set_frame(state+1)
        cmd.refresh()
        if (state+1)%100==0:
            print "wait... %d of %d done..." % (state+1,max_state)
    print "%d of %d done." % (state+1,max_state)
    print "Saving session..."
    cmd.set_frame(1)
    cmd.save(args.session,state=0)
    print "Let the Valve be always open!"
    print "Goodby!"
    cmd.quit()
''')
            self.script_fh.write(os.linesep)
            self.script_fh.close()


class SinglePathPlotter(object):
    def __init__(self, pymol_connector, linearize=None):

        self.cgo_lines = BasicPymolCGOLines()
        self.cgo_spheres = BasicPymolCGOSpheres()
        self.cgo_pointers = BasicPymolCGOPointers()

        # linearize have to be callable! or False or None
        self.linearize = linearize

        self.pymol_connector = pymol_connector

    # TODO: take care of proper colors handling for smoothed and not smoothed traces!

    def add_single_path_continous_trace(self,
                                        spath,
                                        smooth=None,
                                        plot_in=True,
                                        plot_object=True,
                                        plot_out=True,
                                        plot_walk=True,
                                        **kwargs):

        self.cgo_lines.new()

        # get coords
        coords_cont = spath.get_coords_cont(smooth)

        # create slices
        if smooth:
            sls = tuple(list_blocks_to_slices(spath.types_cont))
        else:
            sls = tuple(list_blocks_to_slices(spath.etypes_cont))

        # create traces
        traces_list = tuple([coords_cont[sl] for sl in sls])
        new_line = False
        for trace, sl in zip(traces.midpoints(traces_list), sls):
            if len(trace) > 0:
                # now trace has midpoints
                # get type and etype
                t = spath.types_cont[sl][0]
                et = spath.etypes_cont[sl][0]
                if smooth:
                    et = et[0]
                # plot, if allowed
                if (plot_in and t == PathTypesCodes.path_in_code) or (
                        plot_object and t == PathTypesCodes.path_object_code) or (
                        plot_out and t == PathTypesCodes.path_out_code) or (
                        plot_walk and t == PathTypesCodes.path_walk_code):
                    # get color
                    c = color_codes(et)
                    # now, it is possible to linearize!
                    if smooth and self.linearize:
                        trace = self.linearize(trace)
                    self.cgo_lines.add(trace, cc(c))
                    # new_line = True

            if new_line:
                self.cgo_lines.new()
                new_line = False

    def paths_trace(self, spaths,
                    smooth=None,
                    name='paths',
                    state=None,
                    **kwargs):

        # if state_function is None
        if state is None:
            state = 1
        self.cgo_lines.clean()
        for nr, spath in enumerate(spaths):
            self.add_single_path_continous_trace(spath, smooth=smooth, **kwargs)
        if self.cgo_lines.cgo_entity == self.cgo_lines.cgo_entity_begin:
            self.cgo_lines.clean(empty=True)
        else:
            self.pymol_connector.add_cgo_object(name, self.cgo_lines.get(), state=state)

    def paths_inlets(self, spaths,
                     smooth=None,
                     color=None,
                     plot_in=True,
                     plot_out=True,
                     name='in-out-let',
                     state=None,
                     **kwargs):

        if state is None:
            state = 1
        self.cgo_pointers.clean()
        for nr, spath in enumerate(spaths):
            coords = spath.get_coords_cont(smooth=smooth)
            etypes = spath.etypes_cont
            if plot_in:
                if color is None:
                    c = color_codes(etypes[0])
                else:
                    c = color
                if len(coords) > 1:
                    self.cgo_pointers.add_pointer(point=coords[0],
                                                  direction=coords[1],
                                                  length=1.,
                                                  color=cc(c))
                else:
                    self.cgo_spheres.add(coords=coords, radius=None, color=cc(c))
            if plot_out:
                if color is None:
                    c = color_codes(etypes[-1])
                else:
                    c = color
                if len(coords) > 1:
                    self.cgo_pointers.add_pointer(point=coords[-1],
                                                  direction=coords[-2],
                                                  length=1.,
                                                  color=cc(c),
                                                  reverse=True)
                else:
                    self.cgo_spheres.add(coords=coords, radius=None, color=cc(c))
        self.pymol_connector.add_cgo_object(name, self.cgo_pointers.get(), state=state)

    def scatter(self, coords,
                radius=0.4,
                color='r',
                name='scatter',
                state=None):

        if isinstance(color, str):
            color = cc(color)
        if state is None:
            state = 1

        self.cgo_spheres.clean()

        self.cgo_spheres.add(coords=coords, radius=radius, color=color)

        self.pymol_connector.add_cgo_object(name, self.cgo_spheres.get(), state=state)

    def line(self, coords, color='r', name='line', state=None):

        if isinstance(color, str):
            color = cc(color)
        if state is None:
            state = 1

        self.cgo_lines.clean()

        self.cgo_lines.add(coords=coords, color=color)

        self.pymol_connector.add_cgo_object(name, self.cgo_lines.get(), state=state)

    def multiline_begin(self):
        self.cgo_lines.clean()

    def multiline_add(self, coords, color='r'):
        if isinstance(color, str):
            color = cc(color)
        self.cgo_lines.new()
        self.cgo_lines.add(coords=coords, color=color)

    def multiline_end(self, name=line, state=None):
        if state is None:
            state = 1
        self.pymol_connector.add_cgo_object(name, self.cgo_lines.get(), state=state)

    def convexhull(self, chull,
                   color='m',
                   name='convexhull',
                   state=None):

        if isinstance(color, str):
            color = cc(color)
        if state is None:
            state = 1

        self.cgo_lines.clean()
        for nr, facet in enumerate(chull.facets):
            if nr > 0:
                self.cgo_lines.new()
            self.cgo_lines.add(facet, color=color)

        self.pymol_connector.add_cgo_object(name, self.cgo_lines.get(), state=state)



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2018  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
# Copyright (C) 2019  Tomasz Magdziarz <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module provides functions for clustering.
Clustering is done by :mod:`scikit-learn` module.
"""

from itertools import izip
import numpy as np
from sklearn.cluster import KMeans, MeanShift, estimate_bandwidth
from aquaduct.utils.helpers import Auto
from aquaduct.utils import clui
from aquaduct.traj.barber import WhereToCut
from aquaduct.geom import Sphere


# problems with clustering methods and size of set
# DBSCAN:              n > 0
# AffinityPropagation: n > 0
# KMeans:              n > clusters
# MeanShift:           n > 6


class BarberClusterResult(object):
    """
    Helper class for results of barber clustering.
    """

    def __init__(self, labels_):
        self.labels_ = np.array(labels_)


class BarberCluster(object):
    '''
    Wrapper class that implements *barber* clustering.
    '''

    @staticmethod
    def fit(coords, spheres=None, radii=None):
        '''
        :param Iterable coords: Input coordinates of points to be clustered.
        :param Iterable spheres: Input spheres for each point. Each sphere has center and radius.
        :param Iterable radii: Radii to create spheres using coords.
        '''
        wtc = WhereToCut()
        if spheres is not None:
            wtc.spheres = spheres
        elif radii is not None:
            spheres = [Sphere(center=center, radius=radius, nr=nr) for nr, (center, radius) in
                       enumerate(izip(coords, radii))]
            wtc.spheres = spheres
        else:
            raise TypeError('Either spheres or radii have to be specified.')
        clouds = wtc.cloud_groups(progress=True)
        # clouds to labels!
        labels = np.zeros(len(spheres))
        spheres_id = [s.nr for s in spheres]
        for cloud_id, cloud in clouds.iteritems():
            labels[[spheres_id.index(c) for c in cloud]] = cloud_id
        return BarberClusterResult(labels)


def MeanShiftBandwidth(X, **kwargs):
    '''
    Helper function for automatic calculation of a bandwidth for MeanShift method.

    :param Iterable X: Coordinates of points to be clustered.
    '''
    if 'bandwidth' in kwargs:
        if kwargs['bandwidth'] is Auto:
            bandwidth = estimate_bandwidth(np.array(X),
                                           quantile=0.5)
            # TODO: change it to the default value of 0.3 or use it as option?
            if not bandwidth:
                bandwidth = None
                clui.message("Meanshift automatic bandwidth calculation returned 0; setting bandwidth to None.")
            else:
                clui.message("Meanshift automatic bandwidth calculation: bandwidth = %f" % float(
                    bandwidth))
            kwargs.update({'bandwidth': bandwidth})

    return kwargs


class PerformClustering(object):
    '''
    Helper class for clustering.
    '''

    # aqeuduct clustering helper class

    def __init__(self, method, **kwargs):
        '''
        :param object method: Class that implements cclustering via *fit* method.
        '''

        self.method = method
        self.method_kwargs = kwargs
        self.method_results = None
        self.clusters = None

    def __str__(self):
        out = str(self.method.__name__)
        return out

    def __call__(self, coords, spheres=None):
        # compatibility
        return self.fit(coords, spheres=spheres)

    def _get_noclusters(self, n):
        return [0] * n

    def _get_oneclusters(self, n):
        return [1] * n

    def fit(self, coords, spheres=None):
        '''
        :param Iterable coords: Input coordinates of points to be clustered.
        :param Iterable spheres: Input spheres for each point.
               Optional, important only if :attr:`method` is :class:`BarberCluster`.
        :return: Clusters numbers.
        :rtype: list of int
        '''
        # spheres are used for Barber only
        if len(coords) < 2:
            # single point forms one cluster
            self.clusters = self._get_oneclusters(len(coords))
            return self.clusters
        # special cases
        if self.method is BarberCluster:
            method = self.method()
            self.method_results = method.fit(coords, spheres)
            self.clusters = map(int, self.method_results.labels_ + 1)
        else:
            if self.method is MeanShift:
                if len(coords) < 6:
                    self.clusters = self._get_noclusters(len(coords))
                    clui.message(
                        "Number of objects %d < 6 is too small for MeanShift method. Skipping." % (len(coords)))
                    return self.clusters
                method = self.method(**MeanShiftBandwidth(coords, **self.method_kwargs))
            else:
                if self.method is KMeans:
                    if 'n_clusters' in self.method_kwargs:
                        if len(coords) < self.method_kwargs['n_clusters']:
                            self.clusters = self._get_noclusters(len(coords))
                            clui.message(
                                "Number of objects %d < %d is too small for KMeans method. Skipping." % (
                                    len(coords), self.method_kwargs['n_clusters']))
                            return self.clusters
                method = self.method(**self.method_kwargs)
            self.method_results = method.fit(coords)
            self.clusters = map(int, self.method_results.labels_ + 1)
        return self.clusters

    def centers(self):
        '''
        :return: Centers of clusters.
        '''
        assert self.clusters is not None, "Perform clustering first."
        assert self.method_results is not None, "Perform clustering first."

        if hasattr(self.method_results, 'cluster_centers_'):
            return self.method_results.cluster_centers_
        raise NotImplementedError('Cluster centers is not implemented for %r method yet.' % self.method)



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2018-2019  Tomasz Magdziarz, Michał Banas <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from aquaduct import logger

import re
from collections import OrderedDict, namedtuple
from itertools import imap
from os.path import splitext
from os import pathsep

import numpy as np
import MDAnalysis as mda


def mda_ver():
    return mda.__version__


# FIXME: do it according to user options
if mda.__version__ < '0.16':
    logger.error('Unsupported MDAnalysis version %s; should be 0.16.2 or > 0.19.', mda.__version__)
    raise NotImplementedError('Unsupported MDAnalysis version %s; should be 0.16.2 or > 0.19.' % mda.__version__)

if mda.__version__ >= '0.17' and mda.__version__ < '0.20':
    logger.warning('Unsupported MDAnalysis version %s.', mda.__version__)
    logger.warning('Trying to mitigate potential problems by setting `use_periodic_selections = False`.')
    mda.core.flags['use_periodic_selections'] = False

from MDAnalysis.topology.core import guess_atom_element

from aquaduct.utils.helpers import SmartRange, SmartRangeIncrement
from aquaduct.geom.convexhull import SciPyConvexHull, are_points_within_convexhull
from aquaduct.utils.helpers import arrayify, create_tmpfile, tupleify
from aquaduct.utils.maths import make_default_array
from aquaduct.apps.data import GCS, CRIC
from aquaduct.utils.maths import defaults

################################################################################
# memory decorator

if GCS.cachedir:
    from joblib import Memory

    memory_cache = Memory(cachedir=GCS.cachedir,
                          verbose=0)  # mmap have to be switched off, otherwise smoothing does not work properly
    # memory_cache = Memory(cachedir=GCS.cachedir, mmap_mode='r', verbose=0)
    memory = memory_cache.cache
elif GCS.cachemem:
    from functools import wraps


    # TODO: rework this to adapt it accordingly, moce it to utils?
    # https://medium.com/@nkhaja/memoization-and-decorators-with-python-32f607439f84
    def memory(func):
        cache = func.cache = {}

        @wraps(func)
        def memoized_func(*args, **kwargs):
            key = ','.join(map(str, args)) + '&' + ','.join(map(lambda kv: ':'.join(map(str, kv)), kwargs.iteritems()))
            logger.debug('Looking for cache key %s' % key)
            if key not in cache:
                cache[key] = func(*args, **kwargs)
                logger.debug("New key added to cache.")
            return cache[key]

        return memoized_func
else:
    from aquaduct.utils.helpers import noaction as memory


################################################################################
# trajectory window object

class Window(object):
    def __init__(self, start, stop, step):
        self.start = self._none_or_int(start)
        self.stop = self._none_or_int(stop)
        self.step = self._none_or_int(step)

    @staticmethod
    def _none_or_int(nr):
        if nr is not None:
            return int(nr)

    def __repr__(self):
        return "Window(%r:%r:%r)" % (self.start, self.stop, self.step)

    def correct(self, real_frame_no):
        if self.start is None:
            self.start = 0
        elif self.start < 0:
            self.start = 0
        elif self.start > real_frame_no:
            self.start = real_frame_no - 1

        if self.stop is None:
            self.stop = real_frame_no - 1
        elif self.stop < 0:
            self.stop = 0
        elif self.stop > real_frame_no:
            self.stop = real_frame_no - 1

        if self.step is None:
            self.step = 1
        elif self.step < 0:
            self.step = 1

    def range(self, reverse=False):
        # returns range object
        if reverse:
            return xrange(self.stop, self.start - 1, -1 * self.step)
        return xrange(self.start, self.stop + 1, self.step)

    def get_real(self, frame):
        # recalculates real frame
        return self.start + frame * self.step

    def len(self):
        # lenght of window
        return len(self.range())

    def split(self, slices=None):
        assert slices > 0
        N = int(max(1, np.floor(self.len() / float(slices))))
        for start in list(self.range())[::(N + 1)]:
            yield Window(start, min(self.stop, start + (N) * self.step), self.step)


################################################################################
# MasterReader

# engine problem
# Two options are currently available:
# MDAnalysis
# MDTraj

class OpenReaderTraj(namedtuple('OpenReaderTraj', 'topology trajectory number window engine')):
    def open(self):
        # convenience function
        return open_traj_reader(self)


class MasterReader(object):
    # only one MasterReader object can be used
    # it does not use ANY direct call to ANY MD access software

    open_reader_traj = {}

    topology = ['']
    trajectory = ['']
    window = None  # this is full window

    sandwich_mode = None
    engine_name = 'mda'
    threads = 1
    threads_multiply = 1
    edges = []

    def __del__(self):
        self.reset()

    def __call__(self, topology, trajectory, window=None, sandwich=False, threads=1):
        """
        :param list topology:  List of topologies. Each element is a file name.
        :param list trajectory: List of trajectories. Each element is a file name.
        :param Window window: Frames window to read.
        :param bool sandwich: Flag for setting sandwitch mode.

        If no sandiwch mode is used, number of topologies has to be precisely 1.
        In sandwich mode it can be either 1 or equal to the nuber of trajectory files.
        """

        if not isinstance(topology, list):
            topology = [t.strip() for t in topology.split(pathsep)]
        self.topology = topology
        if not isinstance(trajectory, list):
            trajectory = [t.strip() for t in trajectory.split(pathsep)]
        self.trajectory = trajectory

        self.window = window
        self.sandwich_mode = sandwich

        if len(self.topology) > 1:
            assert self.sandwich_mode, "Multiple topologies possible in sandwich mode only."
            assert len(self.topology) == len(
                self.trajectory), "Number of topologies must be 1 or be equal to number of trajectories."

        self.window.correct(self.real_number_of_frames())  # this corrects window
        self.reset()  # assert window correction clear all opened trajs

        self.threads = threads

    def reset(self):
        self.open_reader_traj = {}

    def __getstate__(self):
        # if pickle dump is required, this will not be used in the future
        # do not pass open_reader_traj
        return dict(((k, v) for k, v in self.__dict__.iteritems() if k not in ['open_reader_traj']))

    def __setstate__(self, state):
        # if pickle dump is required, this will not be used in the future
        self.__dict__ = state
        self.open_reader_traj = {}

    '''
    @property
    def engine(self):
        # returns engine used for accessing trajectory
        # this returns object that inherits from ReaderTraj
        if self.engine_name == 'mda':
            return ReaderTrajViaMDA
        raise NotImplementedError
    '''

    def engine(self, topology, trajectory, number, window):
        '''
        :param str topology: Topology file name.
        :param list trajectory: List of trajectories. Each element is a file name. Alternatively, trajectory file name.
        :param int number: Number of the layer.
        :param Window window: Frames window.
        :return:
        '''
        assert self.engine_name in ['mda']
        return OpenReaderTraj(topology, trajectory, number, window, self.engine_name)

    def __repr__(self):
        sandwich = ''
        if self.sandwich_mode:
            sandwich = ',sandwich'
        return "Reader(%s,%s,%r%s)" % (
        "[%s]" % (','.join(self.topology)), "[%s]" % (','.join(self.trajectory)), self.window, sandwich)

    def sandwich(self, number=False):
        # generates readers of consecutive sandwich layers
        for nr, trajectory in enumerate(self.trajectory):
            if number:
                yield nr, self.get_single_reader(nr)
            else:
                yield self.get_single_reader(nr)

    def baguette(self, number=False):
        # generates reader of trajectory file(s)
        if number:
            yield 0, self.get_single_reader(0)
        else:
            yield self.get_single_reader(0)

    def strata(self, number=False):
        # generates slices of baquette
        for nr in xrange(self.number_of_layers()):
            if number:
                yield nr + 1, self.get_single_reader(nr + 1)
            else:
                yield self.get_single_reader(nr + 1)

    def iterate(self, number=False, threads=True):
        # iterates over trajectory readers
        # calls sandwich or baguette
        if self.sandwich_mode:
            return self.sandwich(number=number)
        elif threads and self.threads > 1:
            return self.strata(number=number)
        return self.baguette(number=number)

    def get_single_raw_reader_per_trajectory(self, number):
        # here number is a number of trajectory, starting from 0
        if len(self.topology) == len(self.trajectory):
            return self.engine(self.topology[number],
                               [self.trajectory[number]],
                               number=number,
                               window=Window(0, None, 1))
        else:
            return self.engine(self.topology[0],
                               [self.trajectory[number]],
                               number=number,
                               window=Window(0, None, 1))

    def get_single_reader(self, number):
        # returns single trajectory reader of number
        # is it is already opened it is returned directly
        # if not is opened and then recursive call is executed
        # print "GetSingleReader(%r)" % (number,)
        if self.sandwich_mode:
            if len(self.topology) == 1:
                return self.engine(self.topology[0], self.trajectory[number], number=number, window=self.window)
            else:
                return self.engine(self.topology[number], self.trajectory[number], number=number, window=self.window)
        elif number > 0 and self.threads > 1:
            window = list(self.window.split(self.number_of_layers()))[number - 1]
            return self.engine(self.topology[0], self.trajectory, number=number, window=window)
        else:
            assert number == 0, "Sandwich/baguette mismatch. Try use/not use --sandwich option."
            return self.engine(self.topology[0], self.trajectory, number=0, window=self.window)

    def get_reader_by_id(self, someid):
        # returns trajectory reader of number in someid
        # someid is tuple (number,ix)
        return self.get_single_reader(someid[0])

    def real_number_of_frames(self):
        # returns number of frames in traj files
        # in sandwich it returns number of frames in first layer
        # in baguette it returns total number of frames (so it is also number of frames in layer 0)
        return open_traj_reader(self.get_single_reader(0)).real_number_of_frames()

    def get_edges(self):
        # should return list of edges of trajectory
        # if one trajectory file is used there are two edges begin and end
        # for each additional trajectory file additional edge marking joining point is returned
        # make sense in non sandwich mode only
        # reader has to be reset before and after
        self.reset()
        # to calculate this each individual file should be open as separate trajectory
        # 1. get real number of frames for each part
        rnf = [0]
        for number in range(len(self.trajectory)):
            # if len(rnf)==0:
            #    rnf.append(0)
            # else:
            #    rnf.append(rnf[-1]+1)
            # rnf.append(rnf[-1]-1+open_traj_reader(self.get_single_raw_reader_per_trajectory(number)).real_number_of_frames())
            rnf.append(open_traj_reader(self.get_single_raw_reader_per_trajectory(number)).real_number_of_frames())
        # cumulative sum
        rnf = (np.cumsum(rnf) - 1).tolist()
        # 2. by using window try to calculate where are the edges
        E = []  # list of edges
        for nr, rf in enumerate(self.window.range()):  # iterate over real frames
            if rf >= rnf[0]:
                # this is an edge
                if self.window.step > 1:
                    E.append(nr - 1)
                else:
                    E.append(nr)
                while len(rnf) and rf >= rnf[0]:
                    rnf.pop(0)
        if len(rnf):
            E.append(nr)
        self.reset()
        return E[1:-1]

    def number_of_frames(self, onelayer=False):
        # number of frames in the window times number of layers
        # in baguette number of layers is 1
        if self.sandwich_mode and not onelayer:
            return len(self.trajectory) * self.window.len()
        return self.window.len()

    def number_of_layers(self):
        if self.sandwich_mode:
            return len(self.trajectory)
        nof = self.number_of_frames()
        nol = self.threads * self.threads_multiply
        return nol if nof > nol else max(nof - 1, 1)


# instance of MasterReader
Reader = MasterReader()


def open_traj_reader_engine(ort):
    if ort.engine == 'mda':
        return ReaderTrajViaMDA(ort.topology, ort.trajectory, number=ort.number, window=ort.window)
    raise NotImplementedError


def open_traj_reader(ort):
    if ort.number not in Reader.open_reader_traj:
        Reader.open_reader_traj.update({ort.number: open_traj_reader_engine(ort)})
    return Reader.open_reader_traj[ort.number]


class ReaderAccess(object):
    # ReaderAccess class provides reader property that returns current instance of MasterReader

    def get_reader(self, number):
        if number in Reader.open_reader_traj:
            # print "Getting reader",number,"from opened readers."
            return Reader.open_reader_traj[number]
        Reader.get_single_reader(number).open()
        return self.get_reader(number)

    def get_reader_by_id(self, someid):
        return self.get_reader(someid[0])

    def get_edges(self):
        return Reader.edges


################################################################################
# VdW radii

VdW_radii = {'ac': 2.47,
             'ag': 2.11,
             'al': 1.84,
             'am': 2.44,
             'ar': 1.88,
             'as': 1.85,
             'at': 2.02,
             'au': 2.14,
             'b': 1.92,
             'ba': 2.68,
             'be': 1.53,
             'bi': 2.07,
             'bk': 2.44,
             'br': 1.85,
             'c': 1.7,
             'ca': 2.31,
             'cd': 2.18,
             'ce': 2.42,
             'cf': 2.45,
             'cl': 1.75,
             'cm': 2.45,
             'co': 2.0,
             'cr': 2.06,
             'cs': 3.43,
             'cu': 1.96,
             'dy': 2.31,
             'er': 2.29,
             'es': 2.45,
             'eu': 2.35,
             'f': 1.47,
             'fe': 2.04,
             'fm': 2.45,
             'fr': 3.48,
             'ga': 1.87,
             'gd': 2.34,
             'ge': 2.11,
             'h': 1.1,
             'he': 1.4,
             'hf': 2.23,
             'hg': 2.23,
             'ho': 2.3,
             'i': 1.98,
             'in': 1.93,
             'ir': 2.13,
             'k': 2.75,
             'kr': 2.02,
             'la': 2.43,
             'li': 1.82,
             'lr': 2.46,
             'lu': 2.24,
             'md': 2.46,
             'mg': 1.73,
             'mn': 2.05,
             'mo': 2.17,
             'n': 1.55,
             'na': 2.27,
             'nb': 2.18,
             'nd': 2.39,
             'ne': 1.54,
             'ni': 1.97,
             'no': 2.46,
             'np': 2.39,
             'o': 1.52,
             'os': 2.16,
             'p': 1.8,
             'pa': 2.43,
             'pb': 2.02,
             'pd': 2.1,
             'pm': 2.38,
             'po': 1.97,
             'pr': 2.4,
             'pt': 2.13,
             'pu': 2.43,
             'ra': 2.83,
             'rb': 3.03,
             're': 2.16,
             'rh': 2.1,
             'rn': 2.2,
             'ru': 2.13,
             's': 1.8,
             'sb': 2.06,
             'sc': 2.15,
             'se': 1.9,
             'si': 2.1,
             'sm': 2.36,
             'sn': 2.17,
             'sr': 2.49,
             'ta': 2.22,
             'tb': 2.33,
             'tc': 2.16,
             'te': 2.06,
             'th': 2.45,
             'ti': 2.11,
             'tl': 1.96,
             'tm': 2.27,
             'u': 2.41,
             'v': 2.07,
             'w': 2.18,
             'xe': 2.16,
             'y': 2.32,
             'yb': 2.26,
             'zn': 2.01,
             'zr': 2.23}
'''
Dictionary of VdW radii.

Data taken from L. M. Mentel, mendeleev, 2014. Available at: https://bitbucket.org/lukaszmentel/mendeleev.
Package **mendeleev** is not used because it depends on too many other libraries.
'''


################################################################################
# Reader Traj base class

# class ReaderTraj(ReaderAccess):
class ReaderTraj(object):
    def __init__(self, topology, trajectory,
                 number=None, window=None):
        """
        :param str topology:  Topology file name.
        :param list trajectory: Trajectory file name.
        :param int number: Number of trajectory file.
        :param Window window: Frames window to read.
        :param Reader reader: Parent reader object.

        This is base class for MD data access engines.
        """

        self.topology = topology  # here, topology is only one file
        self.trajectory = trajectory

        if not isinstance(trajectory, list):
            self.trajectory = [t.strip() for t in trajectory.split(pathsep)]

        # print "ReaderTraj(%r,%r)" % (self.topology,self.trajectory)

        self.number = number
        self.window = window

        self.real_frame_nr = None

        self.trajectory_object = self.open_trajectory()

    def __repr__(self):
        if len(self.trajectory) == 1:
            return "%s(%s,%s,%d,%r)" % (
                self.__class__.__name__, self.topology, self.trajectory[0], self.number, self.window)
        return "%s(%s,%s,%d,%r)" % (
            self.__class__.__name__, self.topology, "[%s]" % (','.join(self.trajectory)), self.number, self.window)

    def iterate_over_frames(self, reverse=False):
        window_len = self.window.len()
        for frame, real_frame in enumerate(self.window.range(reverse=reverse)):
            self.set_real_frame(real_frame)
            if reverse:
                yield window_len - frame - 1
            else:
                yield frame

    def iterate(self, reverse=False):
        return self.iterate_over_frames(reverse=reverse)

    def set_frame(self, frame):
        return self.set_real_frame(self.window.get_real(frame))

    def dump_frames(self, frames, selection=None, filename=None):
        if filename is None:
            filename = create_tmpfile(ext='pdb')
        self.dump_frames_to_file(frames, selection, filename)
        return filename

    def __del__(self):
        self.close_trajectory()

    def number_of_frames(self):
        # should return number of frames in this reader (window)
        return self.window.len()

    def open_trajectory(self):
        # should return any object that can be further used to parse trajectory
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")

    def close_trajectory(self):
        # should close trajectory reader in self.trajectory_object
        # WARNING: This method has to be carefully implemented because it is used by __del__ and
        #          should not emmit any error messages. This is subjet of change.
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")

    def set_real_frame(self, real_frame):
        self.real_frame_nr = real_frame
        # sets real frame
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")

    def real_number_of_frames(self):
        # should return number of frames
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")

    def parse_selection(self, selection):
        # should return selection
        # selection resolution should be atoms
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")

    def select_all(self):
        # should return selection of all atoms
        # selection resolution should be atoms
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")

    def atom_vdw(self, atomid):
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")

    def atom2residue(self, atomid):
        # one atom id to one residue id
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")

    def atoms_positions(self, atomids):
        # atoms ids to coordinates
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")

    def residues_positions(self, resids):
        # residues ids to center of masses coordinates
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")

    def residues_names(self, resids):
        # residues ids to center of masses coordinates
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")

    def atoms_masses(self, atomids):
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")

    def dump_frames_to_file(self, frames, selection, filename):
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")


################################################################################
# ReaderTraj engine MDAnalysis

mda_available_formats = {re.compile('(nc|NC)'): 'nc',
                         re.compile('(prmtop|parmtop|top|PRMTOP|PARMTOP|TOP)'): 'PRMTOP',
                         re.compile('(dcd|DCD)'): 'LAMMPS',
                         re.compile('(psf|PSF)'): 'psf',
                         re.compile('(pdb|PDB)'): 'pdb',
                         re.compile('(crd|CRD)'): 'crd',
                         re.compile('(xtc|XTC)'): 'XTC'}


class ReaderTrajViaMDA(ReaderTraj):

    def open_trajectory(self):
        topology = splitext(self.topology)[1][1:]
        for afk in mda_available_formats.keys():
            if afk.match(topology):
                topology = mda_available_formats[afk]
                break
        trajectory = splitext(self.trajectory[0])[1][1:]
        for afk in mda_available_formats.keys():
            if afk.match(trajectory):
                trajectory = mda_available_formats[afk]
                break
        return mda.Universe(self.topology,
                            self.trajectory,
                            topology_format=topology,
                            format=trajectory)

    def close_trajectory(self):
        if hasattr(self, 'trajectory_object'):
            if hasattr(self.trajectory_object, 'trajectory'):
                if hasattr(self.trajectory_object.trajectory, 'close'):
                    self.trajectory_object.trajectory.close()

    def set_real_frame(self, real_frame):
        self.real_frame_nr = real_frame
        self.trajectory_object.trajectory[real_frame]

    def parse_selection(self, selection):
        return AtomSelection({self.number: self.trajectory_object.select_atoms(selection).atoms.ix})

    def select_all(self):
        return self.parse_selection("name *")

    def atom2residue(self, atomid):
        return self.trajectory_object.atoms[atomid].residue.ix

    def atoms_positions(self, atomids):
        # atoms ids to coordinates
        return self.trajectory_object.atoms[atomids].positions

    def residues_positions(self, resids):
        # residues ids to center of masses coordinates
        return (res.atoms.center_of_geometry() for res in
                self.trajectory_object.residues[list(resids)])

    def residues_names(self, resids):
        # residues ids to center of masses coordinates
        for rid in resids:
            yield self.trajectory_object.residues[rid].resname

    def real_number_of_frames(self):
        # should return number of frames
        return len(self.trajectory_object.trajectory)

    def atoms_masses(self, atomids):
        return self.trajectory_object.atoms[atomids].masses

    def atom_vdw(self, atomid):
        element = str(guess_atom_element(self.trajectory_object.atoms[atomid].name)).lower()
        if element in VdW_radii:
            return VdW_radii[element]
        element = str(self.trajectory_object.atoms[atomid].element).lower()
        if element in VdW_radii:
            return VdW_radii[element]
        return 1.4

    def dump_frames_to_file(self, frames, selection, filename):
        mdawriter = mda.Writer(filename, multiframe=True)
        to_dump = self.trajectory_object.select_atoms(selection)
        for frame in frames:
            self.set_frame(frame)
            mdawriter.write(to_dump)
        mdawriter.close()


################################################################################
# Selection objects

class Selection(ReaderAccess):

    def __init__(self, selected):

        self.selected = OrderedDict(selected)
        for number, ids in self.selected.iteritems():
            self.selected[number] = list(imap(defaults.int_default, ids))

    def layer(self, number):
        if number in self.selected:
            return self.__class__({number: self.selected[number]})
        return self.__class__({})

    def numbers(self):
        return self.selected.keys()

    def ix(self, ix):
        # gets selection of index ix
        ix_current = 0
        for number, ids in self.selected.iteritems():
            if ix_current + len(ids) >= ix + 1:
                # it is here!
                return self.__class__({number: [ids[ix - ix_current]]})  # FIXME: looks like a bug!
            ix_current += len(ids)
        raise IndexError()

    def single(self, selection_id):
        # TODO: suboptimal and probably does not make any sense
        if selection_id[0] in self.selected:
            if selection_id[-1] in self.selected[selection_id[0]]:
                return self.__class__({selection_id[0]: [selection_id[-1]]})
        raise IndexError()

    def len(self):
        _len = 0
        for number, ids in self.selected.iteritems():
            _len += len(ids)
        return _len

    def __len__(self):
        return self.len()

    def add(self, other):

        for number, ids in other.selected.iteritems():
            if number in self.selected:
                self.selected[number] = self.selected[number] + list(ids)
            else:
                self.selected.update({number: ids})

    def remove(self, other):
        """
        Remove all items that exist in other selection.

        :param other: Other selection.
        """
        empty_keys = []
        for number, ids in other.selected.iteritems():
            self.selected[number] = [id_ for id_ in self.selected[number] if id_ not in ids]

            if not self.selected[number]:
                empty_keys.append(number)

        for k in empty_keys:
            del self.selected[k]

    def uniquify(self):

        for number, ids in self.selected.iteritems():
            self.selected[number] = sorted(set(ids))

    def ids(self):
        # report it in this order! always!
        # these are not unique! run run uniqify first!
        for number, ids in self.selected.iteritems():
            for i in ids:
                yield (number, i)

    def coords(self):
        # order of coords should be the same as in ids!
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")

    def center_of_mass(self):
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")


################################################################################

class AtomSelection(Selection):

    def vdw(self):
        for number, ids in self.selected.iteritems():
            for aid in ids:
                yield self.get_reader(number).atom_vdw(aid)

    def residues(self):
        # returns residues selection
        def get_unique_residues():
            for number, ids in self.selected.iteritems():
                number_reader = self.get_reader(number)
                yield number, sorted(set(map(number_reader.atom2residue, ids)))

        return ResidueSelection(get_unique_residues())

    def coords(self):
        for number, ids in self.selected.iteritems():
            number_reader = self.get_reader(number)
            for coord in number_reader.atoms_positions(ids):  # .tolist():
                yield coord

    def center_of_mass(self):
        center = np.zeros(3)
        total_mass = 0
        for number, ids in self.selected.iteritems():
            masses = self.get_reader(number).atoms_masses(ids)
            masses.shape = (len(masses), 1)
            total_mass += sum(masses)
            center += (masses * self.get_reader(number).atoms_positions(ids)).sum(0)
        return center / total_mass

    def contains_residues(self, other_residues, convex_hull=False, convex_hull_inflate=None, map_fun=None,
                          known_true=None):
        # FIXME: known_true slows down!
        # known_true are only ix!
        assert isinstance(other_residues, ResidueSelection)
        if map_fun is None:
            map_fun = map
        # known_true should be list of ids
        if known_true is not None and len(known_true):
            this_ids = list(self.residues().ids())
            kt_list = []
            ch_list = []
            other_coords = []
            for other_id, other_coord in zip(other_residues.ids(), other_residues.coords()):
                if other_id[-1] in known_true:
                    kt_list.append(other_id)
                elif convex_hull:
                    ch_list.append(other_id)
                    other_coords.append(other_coord)
                elif other_id in this_ids:
                    kt_list.append(other_id)
            if convex_hull:
                chull = self.chull(inflate=convex_hull_inflate)
                # ch_results = map_fun(is_point_within_convexhull, izip_longest(other_coords, [], fillvalue=chull))
                ch_results = are_points_within_convexhull(other_coords, chull)
            # final merging loop
            final_results = []
            for other_id in other_residues.ids():
                if other_id in ch_list:
                    final_results.append(ch_results[ch_list.index(other_id)])
                elif other_id in kt_list:
                    final_results.append(True)
                else:
                    final_results.append(False)
            return final_results
        else:
            if convex_hull:
                other_coords = other_residues.coords()
                chull = self.chull(inflate=convex_hull_inflate)
                # return map_fun(is_point_within_convexhull, izip_longest(other_coords, [], fillvalue=chull))
                return are_points_within_convexhull(other_coords, chull)
            else:
                # check if other selection is empty
                this_ids = list(self.residues().ids())
                return [other_id in this_ids for other_id in other_residues.ids()]

    def containing_residues(self, other_residues, *args, **kwargs):
        # Convenience wrapper for contains_residues.
        # def get_res_in_scope(is_res_in_scope, res):
        other_new = OrderedDict()
        for iris, (number, resid) in zip(self.contains_residues(other_residues, *args, **kwargs), other_residues.ids()):
            if iris:
                if number in other_new:
                    other_new[number].append(resid)
                else:
                    other_new.update({number: [resid]})
        return ResidueSelection(other_new)

    def chull(self, inflate=None):
        if self.len() > 3:
            return SciPyConvexHull(list(self.coords()), inflate=inflate)


################################################################################

class ResidueSelection(Selection):

    def coords(self):
        for number, ids in self.selected.iteritems():
            number_reader = self.get_reader(number)
            for coord in number_reader.residues_positions(ids):
                yield coord  # .tolist()

    def names(self):
        for number, ids in self.selected.iteritems():
            for name in self.get_reader(number).residues_names(ids):
                yield name

    def single_residues(self):
        for resid in self.ids():
            yield SingleResidueSelection(resid)


################################################################################

@memory
def coords_range_core(srange, number, rid):
    assert isinstance(srange, SmartRangeIncrement), "Expecting SmartRangeIncrement, got %r" % type(srange)

    @arrayify(shape=(None, 3))
    def coords_range_core_inner(srange, number, rid):
        reader = Reader.get_single_reader(number).open()
        for f in srange.get():
            reader.set_frame(f)
            yield reader.residues_positions([rid]).next()

    return coords_range_core_inner(srange, number, rid)


@arrayify(shape=(None, 3))
def coords_range(srange, number, rid):
    # srange is SmartRangeIncrement, it cannot be anything else
    # wrapper to limit number of calls to coords_range_core
    # CRIC cache is {number:{rid:FramesRangeCollection}}
    logger.debug("CRIC global request %d:%d %s", number, rid, str(srange))
    '''
    if number not in CRIC.cache:
        CRIC.cache.update({number:{}})
        logger.debug("CRIC new number %d",number)
    if rid not in CRIC.cache[number]:
        CRIC.cache[number].update({rid: FramesRangeCollection()})
        logger.debug("CRIC new rid %d",rid)
    '''

    # call get_ranges and stack array, do it in comprehension? nested function?
    def get_coords_from_cache():
        for sr, xr in CRIC.get_frc(number, rid).get_ranges(srange):
            logger.debug("CRIC partial request %d:%d %s", number, rid, str(sr))
            yield coords_range_core(sr, number, rid)[xr, :]

    return np.vstack(tuple(get_coords_from_cache()))


################################################################################

@memory
def smooth_coords_ranges(sranges, number, rid, smooth):
    @tupleify
    def smooth_coords_ranges_inner(sranges, number, rid, smooth):
        # here, sranges are list of SmartRange objects whereas coords_range accepts SmartRangeIncrement
        # first get all coords and make in continous
        def sranges2coords_cont():
            for srange in sranges:
                for srangei in srange.raw:
                    yield coords_range(srangei, number, rid)

        coords_cont = make_default_array(np.vstack([c for c in sranges2coords_cont() if len(c) > 0]))
        # call smooth
        coords_cont = smooth(coords_cont)
        # split coords_cont
        # now lets return tupple of coords
        nr = 0
        for path in sranges:
            if len(path) > 0:
                yield make_default_array(coords_cont[nr:nr + len(path)])
                nr += len(path)
            else:
                # TODO: reuse some kind of empty coords
                yield make_default_array(np.zeros((0, 3)))

    return smooth_coords_ranges_inner(sranges, number, rid, smooth)


################################################################################

class SingleResidueSelection(ReaderAccess):
    def __init__(self, resid):
        super(SingleResidueSelection, self).__init__()
        # where resid is id reported by ResidueSelection and reader is Reader
        # resid is tuple (number,id) number is used to get reader_traj
        self.resid = resid[-1]
        self.number = resid[0]

    def coords(self, frames):
        if isinstance(frames, SmartRange):
            if len(frames):
                return np.vstack([coords_range(srange, self.number, self.resid) for srange in frames.raw])
            return self._coords([])
        else:
            return self._coords(frames)

    @arrayify(shape=(None, 3))
    def _coords(self, frames):
        # return coords for frames
        if len(frames):
            # print "Trying to get reader",self.number,"for",len(frames),"frames"
            traj_reader = self.get_reader(self.number)
            for f in frames:
                traj_reader.set_frame(f)
                yield traj_reader.residues_positions([self.resid]).next()

    def coords_smooth(self, sranges, smooth):
        for coord in smooth_coords_ranges(sranges, self.number, self.resid, smooth):
            yield coord


################################################################################

# flag sandwich as imported
GCS.sandwich_import = True



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2018-2019  Michał Banas
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import Tkinter as tk
import os
import ttk
from tkFileDialog import askopenfile, askdirectory

import defaults
from aquaduct.apps.valveconfig import get_img


def get_widget_bg(widget):
    """
    Return background color of specified widget.

    :param widget: Ttk widget.
    :return: Background color.
    :rtype: str
    """
    try:
        return ttk.Style().lookup(widget["style"], "background")
    except tk.TclError:
        # FIXME: Appending T may cause errors
        return ttk.Style().lookup("T" + widget.winfo_class(), "background")


def widget_factory(parent, default, state=tk.NORMAL):
    """
    Creates widget depending on default argument.

    :param parent: Parent of new widget.
    :param default: Default widget value.
    :param state: State of widget.
    :return: Widget and variable attached to it.
    :rtype: tuple
    """
    if isinstance(default, defaults.longstr):  # Due to inheritance from str, longstr must be checked first
        v = tk.StringVar()
        v.set(default)

        w = Text(parent, textvariable=v, wrap=tk.WORD, state=state, width=34, height=5)
    elif isinstance(default, str):
        v = tk.StringVar()
        v.set(default)

        w = ttk.Entry(parent, textvariable=v, width=30, state=state, background=get_widget_bg(parent))
    elif isinstance(default, bool):
        v = tk.BooleanVar()
        v.set(default)

        w = ttk.Checkbutton(parent, variable=v, offvalue=False, onvalue=True, state=state)
    elif isinstance(default, int):
        v = tk.IntVar()
        v.set(default)

        w = ttk.Entry(parent, textvariable=v, width=5, state=state, background=get_widget_bg(parent))
    elif isinstance(default, float):
        v = tk.DoubleVar()
        v.set(default)

        w = ttk.Entry(parent, textvariable=v, width=5, state=state, background=get_widget_bg(parent))
    elif isinstance(default, tuple):
        v = tk.StringVar()

        w = ttk.OptionMenu(parent, v, default[0], *default)
    elif isinstance(default, list):
        v = tk.StringVar()

        w = ttk.Combobox(parent, textvariable=v)
        w["values"] = default
        w.current(0)
    else:
        raise TypeError("There is no specified behaviour for {} type".format(type(default)))

    return w, v


def entry_factory(parent, row, entry_name, default, help, state=tk.NORMAL, info_text=None, warning_text=None):
    """
    Determines which class is used to handle specified default value.

    :param parent: Parent of widget.
    :param row: Row number where first Entry will be grided.
    :param entry_name: Readable entry name.
    :param default: Default values of entry.
    :param help: Text which will be displayed in tooltip.
    :param state: State of widget.
    :return: Entry based on default value.
    """
    if len(default) > 2:
        raise RuntimeError("There can be only two values in config defaults({})".format(entry_name))
    elif len(default) == 2:
        input_default, control_default = default

        if isinstance(control_default, bool):
            return BoolEntry(parent, row, entry_name, input_default, control_default, help, info_text, warning_text)
        elif isinstance(control_default, float):
            return ParenthesedEntry(parent, row, entry_name, input_default, control_default, help, info_text,
                                    warning_text)
        else:
            raise TypeError("There is no specified behaviour for {} type(for {} option). "
                            "First must be input widget, then control widget"
                            .format(type(control_default), entry_name))
    elif isinstance(default[0], defaults.filetype):
        return FileEntry(parent, row, entry_name, str(), help, info_text, warning_text)
    elif isinstance(default[0], defaults.manyfiletype):
        return ManyFileEntry(parent, row, entry_name, str(), help, info_text, warning_text)
    elif isinstance(default[0], defaults.dirtype):
        return DirEntry(parent, row, entry_name, default[0], help, info_text, warning_text)

    else:
        return StandardEntry(parent, row, entry_name, default[0], help, state, info_text, warning_text)


class Text(tk.Text, object):
    def __init__(self, parent, textvariable, **kwargs):
        """
         Text widget with ability to assign content to variable.

        :param parent: Parent of widget.
        :param textvariable: String variable to which will contain Text content.
        :param kwargs: Arguments which will be passed to original tk.Text widget.
        """
        super(Text, self).__init__(parent, **kwargs)

        self.var = textvariable

        # Block performing callback recursivelly
        self.block = False

        if self.var:
            # Every written char assign content to var
            self.bind("<Key>", self._on_update)

        if self.var:
            # Change Text content when variable will change
            self.var.trace("w", self._on_var_update)

    def _on_update(self, e):
        """
        Text widget callback which assign its content to variable.

        :param e: Event informations.
        """
        if not self.block:
            self.block = True

            self.var.set(self.get("1.0", tk.END))

            self.block = False

    def _on_var_update(self, *args):
        """
        Variable callback which assign its content to Text widget.

        :param args: Event informations.
        """
        if not self.block:
            self.block = True

            self.delete('1.0', tk.END)
            self.insert("1.0", self.var.get())

            self.block = False


class Entry(object):
    """
    Abstract class for various Entries that manage different type of default values.

    Represents single row of configuration option with Label and all input widgets.
    """

    def __init__(self, parent, row):
        self.input_var = None
        self.control_var = None

        self.input_frame = tk.Frame(parent)
        self.input_frame.grid(row=row, column=1, sticky="w", pady=5)

        self.hightlight_color = "OrangeRed2"
        self.default_background = None

        self.label_sticky = "E"

    def get(self):
        """
        Gets Entry value.

        :return: Entry value.
        """
        raise NotImplementedError()

    def set(self, value):
        """
        Sets Entry value.

        :param value: New value of Entry.
        """
        raise NotImplementedError()

    def highlight(self):
        """
        Changes color of input frame.

        Used to highlight, which required entry is unfilled.
        """
        self.default_background = self.input_frame.cget("background")
        self.input_frame.config(background=self.hightlight_color)

    def unhighlight(self):
        """
        Sets entry to default color.
        """
        self.input_frame.configure(bg=self.default_background)


class StandardEntry(Entry):
    def __init__(self, parent, row, entry_name_long, default, help, state, info_text=None, warning_text=None):
        """
        Entry with standard widget.

        :param parent: Parent of widgets.
        :param row: Row where widgets will be grided.
        :param entry_name_long: Readable entry name.
        :param default: Default values of entry.
        :param help: Text which will be displayed in tooltip.
        :param state: State of widget.
        """
        super(StandardEntry, self).__init__(parent, row)

        ttk.Label(parent, text=entry_name_long, background=get_widget_bg(parent)).grid(sticky=self.label_sticky,
                                                                                       row=row, column=0)

        widget, self.input_var = widget_factory(self.input_frame, default, state)
        widget.pack(side=tk.LEFT, padx=5, pady=5)

        if info_text:
            InfoIconWidget(self.input_frame, info_text).pack(side=tk.LEFT)
        elif warning_text:
            WarningIconWidget(self.input_frame, warning_text).pack(side=tk.LEFT)

        ToolTip.create(widget, help)

    def get(self):
        """
        Gets Entry value.

        :return: Entry value.
        """
        return self.input_var.get()

    def set(self, value):
        """
        Sets Entry value.

        :param value: New value of Entry.
        """
        self.input_var.set(value)


class BoolEntry(Entry):
    def __init__(self, parent, row, entry_name_long, input_default, control_default, help, info_text=None,
                 warning_text=None):
        """
        Entry with Checkbox and Entry or text widget.

        :param parent: Parent of widgets.
        :param row: Row where widgets will be grided.
        :param entry_name_long: Readable entry name.
        :param default: Default values of entry.
        :param help: Text which will be displayed in tooltip.
        :param state: State of widget.
        """
        super(BoolEntry, self).__init__(parent, row)

        self.entry_name_long = entry_name_long

        ttk.Label(parent, text=entry_name_long).grid(sticky=self.label_sticky, row=row, column=0)

        input_widget, self.input_var = widget_factory(self.input_frame, input_default)
        input_widget.pack(side=tk.RIGHT)

        ToolTip.create(input_widget, help)

        control_widget, self.control_var = widget_factory(self.input_frame, control_default)
        control_widget.pack(side=tk.LEFT)

        if info_text:
            InfoIconWidget(self.input_frame, info_text).pack(side=tk.LEFT)
        elif warning_text:
            WarningIconWidget(self.input_frame, warning_text).pack(side=tk.LEFT)

        ToolTip.create(control_widget, help)

    def get(self):
        """
        Get Entry value.

        :return: If Checkbox is checked it return input widget value, otherwise False.
        """
        if self.control_var.get():
            return self.input_var.get()
        else:
            return False

    def set(self, value):
        """
        Set Entry value.

        :param value: Value that input widget will be set too.
        If False it will set only Checkbox value to False and input widget to "".
        If value have str type it will set Checkbox to True and input widget to that value.
        """
        if value == "False" or value == False:  # If value = 0 expression should be true
            self.control_var.set(False)
            self.input_var.set("")
        else:
            self.control_var.set(True)
            self.input_var.set(value)


class FileEntry(Entry):
    def __init__(self, parent, row, entry_name_long, default, help, info_text=None, warning_text=None):
        """
        Entry with Entry widget and button to load and append file name to it.

        :param parent: Parent of widgets.
        :param row: Row where widgets will be grided.
        :param entry_name_long: Readable entry name.
        :param default: Default values of entry.
        :param help: Text which will be displayed in tooltip.
        :param state: State of widget.
        """
        super(FileEntry, self).__init__(parent, row)

        ttk.Label(parent, text=entry_name_long).grid(sticky=self.label_sticky, row=row, column=0)

        self.input_widget, self.input_var = widget_factory(self.input_frame, default)
        self.input_widget.pack(side=tk.LEFT, padx=5, pady=5)

        ToolTip.create(self.input_widget, help)

        load_file_button = ttk.Button(self.input_frame, text="Load", style="File.TButton")
        load_file_button.pack(side=tk.LEFT, padx=5)

        load_file_button.bind("<Button-1>", self.callback_load_file)

        if info_text:
            InfoIconWidget(self.input_frame, info_text).pack(side=tk.LEFT)
        elif warning_text:
            WarningIconWidget(self.input_frame, warning_text).pack(side=tk.LEFT)

    def callback_load_file(self, e):
        """
        Callback for selecting file.

        Sets widget content to loaded file name.
        """
        try:
            with askopenfile("r") as f:
                self.input_var.set(f.name)
        except AttributeError:  # In case of cancel selecting file
            pass

    def get(self):
        """
        Gets Entry value.

        :return: Entry value.
        """
        return self.input_var.get()

    def set(self, value):
        """
        Sets Entry value.

        :param value: New value of Entry.
        """
        self.input_var.set(value)


class ManyFileEntry(Entry):
    def __init__(self, parent, row, entry_name_long, default, help, info_text=None, warning_text=None):
        """
        Entry with Text widget and button to load and append file names to it.

        :param parent: Parent of widgets.
        :param row: Row where widgets will be grided.
        :param entry_name_long: Readable entry name.
        :param default: Default values of entry.
        :param help: Text which will be displayed in tooltip.
        :param state: State of widget.
        """
        super(ManyFileEntry, self).__init__(parent, row)

        self.parent = parent
        self.row = row
        self.entry_name_long = entry_name_long
        self.default = default
        self.help = help

        self.input_vars = []
        self.frames = []  # Store single row of widgets

        ttk.Label(parent, text=entry_name_long).grid(sticky=self.label_sticky, row=row, column=0)

        self.append_entry()

        if info_text:
            InfoIconWidget(self.input_frame, info_text).grid(row=0, column=2)
        elif warning_text:
            WarningIconWidget(self.input_frame, warning_text).grid(row=0, column=2)

    def append_entry(self):
        """ Creates new entry with input widget and load button """
        frame = tk.Frame(self.input_frame)
        frame.pack(anchor="nw", ipady=5)
        self.frames.append(frame)

        input_widget, input_var = widget_factory(frame, self.default)
        input_widget.pack(side=tk.LEFT, padx=5)

        self.input_vars.append(input_var)

        ToolTip.create(input_widget, self.help)

        load_file_button = ttk.Button(frame, text="Load", style="File.TButton")
        load_file_button.pack(side=tk.LEFT, padx=5)

        callback = CallbackWrapper(self.callback_load_file, self.input_vars.index(input_var))
        load_file_button.bind("<Button-1>", callback)

    def callback_load_file(self, index):
        """
        Callback for selecting file.

        Appends loaded file name at the end of Text widget.

        :param index: Index of variable in self.input_vars
        """
        try:
            with askopenfile("r") as f:
                self.input_vars[index].set(f.name)

                # Append new entry only if file is loaded to last entry
                if len(self.input_vars) - 1 <= index:
                    self.append_entry()
        except AttributeError:  # In case of cancel selecting file
            pass

    def get(self):
        """
        Gets Entry value.

        :return: Entry value.
        """
        # TODO: validate path separator
        return os.pathsep.join([var.get() for var in self.input_vars if var.get()])

    def set(self, value):
        """
        Sets Entry value.

        If value is set to "" it deletes all input widgets, except first and sets it to ""

        :param value: New value. It can be single path or paths separated by os.pathsep.
        """
        if value == "":
            self.input_vars[0].set("")

            for i in reversed(range(1, len(self.frames))):
                self.frames[i].pack_forget()
                del self.frames[i]
                del self.input_vars[i]

            return

        for i, path in enumerate(value.split(os.pathsep)):
            self.input_vars[i].set(path)
            self.append_entry()

    def highlight(self):
        self.default_background = self.input_frame.cget("background")
        for frame in self.frames:
            frame.config(background=self.hightlight_color)

    def unhighlight(self):
        for frame in self.frames:
            frame.config(background=self.default_background)


class DirEntry(Entry):
    def __init__(self, parent, row, entry_name_long, default, help, info_text=None, warning_text=None):
        """
        Entry with Entry widget and button to load and append file name to it.

        :param parent: Parent of widgets.
        :param row: Row where widgets will be grided.
        :param entry_name_long: Readable entry name.
        :param default: Default values of entry.
        :param help: Text which will be displayed in tooltip.
        :param state: State of widget.
        """
        super(DirEntry, self).__init__(parent, row)

        ttk.Label(parent, text=entry_name_long).grid(sticky=self.label_sticky, row=row, column=0)

        self.input_widget, self.input_var = widget_factory(self.input_frame, default)
        self.input_widget.pack(side=tk.LEFT, padx=5, pady=5)

        ToolTip.create(self.input_widget, help)

        load_file_button = ttk.Button(self.input_frame, text="Load", style="File.TButton")
        load_file_button.pack(side=tk.LEFT, padx=5)

        load_file_button.bind("<Button-1>", self.callback_load_dir)

        if info_text:
            InfoIconWidget(self.input_frame, info_text).pack(side=tk.LEFT)
        elif warning_text:
            WarningIconWidget(self.input_frame, warning_text).pack(side=tk.LEFT)

    def callback_load_dir(self, e):
        """
        Callback for selecting dir.

        Sets widget content to loaded dir name.
        """
        try:
            selected_dir = askdirectory()
            self.input_var.set(selected_dir)
        except AttributeError:  # In case of cancel selecting dir
            pass

    def get(self):
        """
        Gets Entry value.

        :return: Entry value.
        """
        return self.input_var.get()

    def set(self, value):
        """
        Sets Entry value.

        :param value: New value of Entry.
        """
        self.input_var.set(value)


class ParenthesedEntry(Entry):
    def __init__(self, parent, row, entry_name_long, input_default, control_default, help, info_text=None,
                 warning_text=None):
        """
        Entry with Text widget and button to load and append file names to it.

        :param parent: Parent of widgets.
        :param row: Row where widgets will be grided.
        :param entry_name_long: Readable entry name.
        :param default: Default values of entry.
        :param help: Text which will be displayed in tooltip.
        :param state: State of widget.
        """
        super(ParenthesedEntry, self).__init__(parent, row)

        ttk.Label(parent, text=entry_name_long).grid(sticky=self.label_sticky, row=row, column=0)

        input_widget, self.input_var = widget_factory(self.input_frame, input_default)
        input_widget.pack(side=tk.LEFT)

        ToolTip.create(input_widget, help)

        control_widget, self.control_var = widget_factory(self.input_frame, control_default)
        control_widget.pack(side=tk.LEFT)

        if info_text:
            InfoIconWidget(self.input_frame, info_text).pack(side=tk.LEFT)
        elif warning_text:
            WarningIconWidget(self.input_frame, warning_text).pack(side=tk.LEFT)

        ToolTip.create(control_widget, help)

    def get(self):
        """
        Gets Entry value.

        :return: Value of Entry widget and value of second Entry in parentheses, eg. Value1(Value2).
        """
        formatter_string = "{}"
        if self.control_var.get():
            formatter_string = "{}({})"

        return formatter_string.format(self.input_var.get(), self.control_var.get())

    def set(self, value):
        """
        Sets Entry value.

        :param value: First value and second value in parentheses or without second value and parentheses.
        """
        try:
            b_pos1 = value.index("(")
            b_pos2 = value.index(")")
            self.input_var.set(value[0:b_pos1])
            self.control_var.set(value[b_pos1 + 1:b_pos2])
        except ValueError:  # When there is no threshold
            self.input_var.set(value)
            self.control_var.set(0.0)


class WarningIconWidget(ttk.Label, object):
    def __init__(self, parent, text):
        """
        Widget with waring icon and Tooltip information

        :param parent: Parent of widget.
        :param text: Content of tooltip
        """
        self.image = tk.PhotoImage(file=get_img("warning.gif"))
        super(WarningIconWidget, self).__init__(parent, image=self.image, padding=0)

        ToolTip.create(self, text)


class InfoIconWidget(ttk.Label, object):
    def __init__(self, parent, text):
        """
        Widget with info icon and Tooltip information

        :param parent: Parent of widget.
        :param text: Content of tooltip
        """
        self.image = tk.PhotoImage(file=get_img("info.gif"))
        super(InfoIconWidget, self).__init__(parent, image=self.image, padding=0)

        ToolTip.create(self, text)


class HidingFrame(ttk.Frame, object):
    def __init__(self, parent, row, text, **kwargs):
        """
        Frame that remembers inner row for griding new widgets.
        Used to keep methods that depends on option menu value.

        :param parent: Parent of widget.
        :param row: Row where widgets will be grided.
        :param text: Title of the Frame.
        :param kwargs: Arguments which will be passed to original ttk.Frame widget.
        """
        super(HidingFrame, self).__init__(parent, **kwargs)

        ttk.Label(self, text=text, style="HF.TFrame.Label").grid(
            sticky="W", row=0, column=0, columnspan=2, padx=5, pady=5)

        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)

        self.row = row

    def show(self):
        """ Method to grid Frame with predefinied configuration. """
        self.grid(row=self.row, column=0, columnspan=2, padx=10, pady=10, ipady=5, ipadx=10)


class CallbackWrapper(object):
    def __init__(self, callback, *args, **kwargs):
        """ Allow to use callbacks with predefined list of arguments. """
        self.callback = callback
        self.args = args
        self.kwargs = kwargs

    def __call__(self, *args, **kwargs):
        self.callback(*self.args, **self.kwargs)


class ToolTip(object):
    """
    Code found: http://www.voidspace.org.uk/python/weblog/arch_d7_2006_07_01.shtml
    """

    def __init__(self, widget):
        self.widget = widget
        self.tipwindow = None
        self.id = None
        self.x = self.y = 0

    @staticmethod
    def create(widget, text):
        toolTip = ToolTip(widget)

        def enter(event):
            toolTip.showtip(text)

        def leave(event):
            toolTip.hidetip()

        widget.bind('<Enter>', enter)
        widget.bind('<Leave>', leave)

    def showtip(self, text):
        # Display text in tooltip window
        self.text = text
        if self.tipwindow or not self.text:
            return
        x, y, cx, cy = self.widget.bbox("insert")
        x = x + self.widget.winfo_rootx() + 15
        y = y + cy + self.widget.winfo_rooty() + 15
        self.tipwindow = tw = tk.Toplevel(self.widget)
        tw.wm_overrideredirect(1)
        tw.wm_geometry("+%d+%d" % (x, y))
        try:
            # For Mac OS
            tw.tk.call("::tk::unsupported::MacWindowStyle",
                       "style", tw._w,
                       "help", "noActivates")
        except tk.TclError:
            pass
        label = ttk.Label(tw, text=self.text, justify=tk.LEFT,
                          background="#ffffe0", relief=tk.SOLID, borderwidth=1,
                          font=("TkDefaultFont", "8", "normal"), wraplength=400)
        label.pack(ipadx=1)

    def hidetip(self):
        tw = self.tipwindow
        self.tipwindow = None
        if tw:
            tw.destroy()


# http://tkinter.unpythonic.net/wiki/VerticalScrolledFrame
class VerticalScrolledFrame(tk.Frame):
    """A pure Tkinter scrollable frame that actually works!
    * Use the 'interior' attribute to place widgets inside the scrollable frame
    * Construct and pack/place/grid normally
    * This frame only allows vertical scrolling
    """

    def __init__(self, parent, *args, **kw):
        tk.Frame.__init__(self, parent, *args, **kw)

        # create a canvas object and a vertical scrollbar for scrolling it
        vscrollbar = tk.Scrollbar(self, orient=tk.VERTICAL)
        vscrollbar.pack(fill=tk.Y, side=tk.RIGHT, expand=tk.FALSE)
        self.canvas = tk.Canvas(self, bd=0, highlightthickness=0,
                                yscrollcommand=vscrollbar.set)
        self.canvas.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.TRUE)
        vscrollbar.config(command=self.canvas.yview)

        # reset the view
        self.canvas.xview_moveto(0)
        self.canvas.yview_moveto(0)

        # create a frame inside the canvas which will be scrolled with it
        self.interior = interior = tk.Frame(self.canvas)
        interior_id = self.canvas.create_window(0, 0, window=interior, anchor="nw")

        # track changes to the canvas and frame width and sync them,
        # also updating the scrollbar
        def _configure_interior(e):
            # update the scrollbars to match the size of the inner frame
            size = (interior.winfo_reqwidth(), interior.winfo_reqheight())
            self.canvas.config(scrollregion="0 0 %s %s" % size)
            if interior.winfo_reqwidth() != self.canvas.winfo_width():
                # update the canvas's width to fit the inner frame
                self.canvas.config(width=interior.winfo_reqwidth())

        interior.bind('<Configure>', _configure_interior)

        def _configure_canvas(e):
            if interior.winfo_reqwidth() != self.canvas.winfo_width():
                # update the inner frame's width to fill the canvas
                self.canvas.itemconfigure(interior_id, width=self.canvas.winfo_width())

        self.canvas.bind('<Configure>', _configure_canvas)

        self.canvas.bind('<Enter>', self._bind_mousewheel)
        self.canvas.bind('<Leave>', self._unbind_mousewheel)

    def _bind_mousewheel(self, e):
        def _mousewheel_handler(e, canvas):
            if e.num == 5:
                canvas.yview_scroll(1, "units")
            if e.num == 4:
                canvas.yview_scroll(-1, "units")

        self.canvas.bind_all("<MouseWheel>", lambda e: _mousewheel_handler(e, self.canvas))
        self.canvas.bind_all("<Button-4>", lambda e: _mousewheel_handler(e, self.canvas))
        self.canvas.bind_all("<Button-5>", lambda e: _mousewheel_handler(e, self.canvas))

    def _unbind_mousewheel(self, e):
        self.canvas.unbind("<MouseWheel>")



#!/usr/bin/env python2.7
# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2018-2019  Michał Banas
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import Tkinter as tk
import base64
import csv
import re
import tkMessageBox
import ttk
from cStringIO import StringIO
from collections import defaultdict
from tkFileDialog import askopenfile

import matplotlib.pyplot as plt

import aquaduct
import aquaduct.apps.valveconfig.utils as utils
from aquaduct.apps.chord import Chord, color_gen
from aquaduct.apps.valveconfig import get_img


def is_float(value):
    try:
        float(value)
        return True
    except:
        return False


class DataException(Exception):
    pass


class FileDataProcessor(object):
    def __init__(self, filename):
        self.file = open(filename, "r")
        self.file_lines = [line.strip() for line in self.file.readlines()]

        self.table_end_pattern = "-+"

    def get_column_values(self, table_name, column_name):
        data = [row[column_name] for row in self._find_table(table_name)]

        return data

    def get_column_names(self, table_name):
        data = self._find_table(table_name)

        if len(data):
            return list(data[0].iterkeys())
        else:
            raise DataException("Table \"{}\" is empty.".format(table_name))

    def _find_table(self, table_name):
        for i, line in enumerate(self.file_lines):
            if line.strip() == table_name:
                return self._parse_column(i)

        raise DataException("Table \"{}\" does not exist.".format(table_name))

    def _parse_column(self, start_line):
        data = []

        column_names = self.file_lines[start_line + 2].split()

        for line in self.file_lines[start_line + 4:]:
            if re.match(self.table_end_pattern, line):
                break

            values = []
            for value in line.split():
                if value.isdigit():
                    value = int(value)
                elif is_float(value):
                    value = float(value)

                values.append(value)

            dict_values = {}
            for i in range(0, len(values)):
                dict_values.update({column_names[i]: values[i]})

            data.append(dict_values)

        return data

    def __del__(self):
        self.file.close()


class CSVDataProcessor(object):
    def __init__(self, filename):
        self.file = open(filename, "r")
        self.csv_reader = csv.DictReader(self.file)

        self.column_names = list(self.csv_reader.next().iterkeys())
        self.seek_file()

    def get_column_values(self, column_name):
        values = [float(row[column_name]) if is_float(row[column_name]) else int(row[column_name]) for row in
                  self.csv_reader]
        self.seek_file()

        return values

    def seek_file(self):
        """ Set file position at beginning and skip first row with column names """
        self.file.seek(0)
        self.csv_reader.next()

    def __del__(self):
        self.file.close()


# 1
def cluster_inlets(file_processor, suffix=""):
    fig, ax = plt.subplots()

    cluster_no = file_processor.get_column_values("Clusters summary - inlets" + suffix, "Cluster")

    y_in = file_processor.get_column_values("Clusters summary - inlets" + suffix, "INCOMING")
    y_out = file_processor.get_column_values("Clusters summary - inlets" + suffix, "OUTGOING")

    # Added 1 to make place for previous bar
    rects_bottom = ax.bar(range(1, len(cluster_no) + 1), y_in, label="INCOMING")
    rects_top = ax.bar(range(1, len(cluster_no) + 1), y_out, bottom=y_in, label="OUTGOING")

    y_lim = ax.get_ylim()[1]

    for rect_b, rect_t in zip(rects_bottom, rects_top):
        width_b = rect_b.get_width()
        height_b = rect_b.get_height()
        if height_b / y_lim >= 0.02:
            ax.text(rect_b.get_x() + width_b / 2., height_b / 2., height_b,
                    fontsize=7,
                    color=(1, 1, 1),
                    horizontalalignment="center",
                    verticalalignment="center")

        width_t = rect_t.get_width()
        height_t = rect_t.get_height()
        if height_t / y_lim >= 0.02:
            ax.text(rect_t.get_x() + width_t / 2., height_b + height_t / 2., height_t,
                    fontsize=7,
                    color=(1, 1, 1),
                    horizontalalignment="center",
                    verticalalignment="center")

    ax.set_title("Clusters size" + suffix)
    ax.set_xlabel("Cluster")
    ax.set_ylabel("Size")

    ax.set_xlim((0, len(cluster_no) + 1))
    ax.set_xticks(range(1, len(cluster_no) + 1))
    ax.set_xticklabels(cluster_no)
    ax.yaxis.set_major_locator(plt.MaxNLocator(integer=True))

    ax.legend(bbox_to_anchor=(1.02, 1), loc="upper left")

    return fig


# 2
def relative_clusters_flows(csv_processor, clusters_names, labels, colors):
    fig, ax = plt.subplots()

    x = csv_processor.get_column_values("# frame")

    all_paths_per_frame = csv_processor.get_column_values("amol_apaths_aclusts_walk")

    for cluster_name, label, color in zip(clusters_names, labels, colors):
        cluster_paths_per_frame = csv_processor.get_column_values(cluster_name)

        y = [float(cluster_paths) / all_paths if all_paths else 0
             for cluster_paths, all_paths in zip(cluster_paths_per_frame, all_paths_per_frame)]

        ax.plot(x, y, label=label, color=color)

    ax.set_title("Relative cluster flows")
    ax.set_xlabel("Frame")
    ax.set_ylabel("Relative flow")

    ax.set_xlim((0, len(x)))
    ax.set_ylim((0, 1.05))  # added 0.05 because when there is 1, its not visible

    ax.legend(bbox_to_anchor=(1.02, 1), loc="upper left")

    return fig


# 3
def ligands_time(file_processor, molecule=None):
    fig, ax = plt.subplots()

    plot_settings = dict(align="edge", height=1.0)

    title_suffix = "" if not molecule else " of {}".format(molecule)

    path_nr = file_processor.get_column_values("List of separate paths and properties", "Nr")

    path_nr = [y - 1 for y in path_nr]

    beg_values = file_processor.get_column_values("List of separate paths and properties", "BeginF")
    inp_values = file_processor.get_column_values("List of separate paths and properties", "InpF")
    obj_values = file_processor.get_column_values("List of separate paths and properties", "ObjF")
    out_values = file_processor.get_column_values("List of separate paths and properties", "OutF")

    if molecule:
        residues_names = file_processor.get_column_values("List of separate paths and properties", "RES")

        for i, res_name in reversed(list(enumerate(residues_names))):
            if res_name != molecule:
                del path_nr[i]
                del beg_values[i]
                del inp_values[i]
                del obj_values[i]
                del out_values[i]

    # Sort all basing on BeginF
    beg_values, inp_values, obj_values, out_values = zip(*sorted(zip(beg_values,
                                                                     inp_values,
                                                                     obj_values,
                                                                     out_values)))

    inps_left = beg_values
    inps_width = [width if width - 1 >= 0 else 0 for width in inp_values]

    objs_left = [sum(values) for values in zip(beg_values, inp_values)]
    objs_width = [width - 1 for width in obj_values]

    outs_left = [sum(values) for values in zip(beg_values, inp_values, obj_values)]
    outs_width = [width - 1 for width in out_values]

    ax.barh(path_nr, inps_width, left=inps_left, color=(1, 0, 0), label="Incoming", **plot_settings)
    ax.barh(path_nr, objs_width, left=objs_left, color=(0, 1, 0), label="Object", **plot_settings)
    ax.barh(path_nr, outs_width, left=outs_left, color=(0, 0, 1), label="Outgoing", **plot_settings)

    ax.set_title("Molecule Entry Time Distribution" + title_suffix)
    ax.set_xlabel("Frame")
    ax.set_ylabel("Separate path ID")

    ax.set_ylim((0, max(path_nr)))
    ax.yaxis.set_major_locator(plt.MaxNLocator(integer=True))

    ax.legend(bbox_to_anchor=(1.02, 1), loc="upper left")

    return fig


# 4
def chord_diagram_sizes(file_processor, labels={}, colors={}):
    fig, ax = plt.subplots(subplot_kw={"aspect": 1})
    ax.set_title("Cluster size")
    ax.set_axis_off()
    ax.set_xlim(-110, 110)
    ax.set_ylim(-110, 110)

    clusters_ids = file_processor.get_column_values("Clusters summary - inlets", "Cluster")
    clusters_ids = [str(id_) for id_ in clusters_ids]
    clusters_sizes = file_processor.get_column_values("Clusters summary - inlets", "Size")

    # TODO:  Move it to generate() method
    if not labels:
        labels = clusters_ids
    else:
        tmp_labels = []
        for id_ in clusters_ids:
            tmp_labels.append(labels[id_])

        labels = tmp_labels

    if colors:
        tmp_colors = []
        for id_ in clusters_ids:
            tmp_colors.append(colors[id_])

        colors = tmp_colors

    Chord(ax, 100, clusters_sizes, [], labels, colors)

    return fig


# 4
def chord_diagram_flows(file_processor, labels={}, colors={}, threshold=0.):
    fig, ax = plt.subplots(subplot_kw={"aspect": 1})
    ax.set_title("Intramolecular flows")
    ax.set_axis_off()
    ax.set_xlim(-110, 110)
    ax.set_ylim(-110, 110)

    flows = file_processor.get_column_values("Separate paths clusters types summary - mean lengths of paths", "CType")
    flows = [{flow.split(":")[0]: flow.split(":")[1]} for flow in flows]

    flows_sizes = file_processor.get_column_values("Separate paths clusters types summary - mean lengths of paths",
                                                   "Size")

    sizes_d = defaultdict(int)
    for flow, size in zip(flows, flows_sizes):
        source, dest = next(flow.iteritems())

        if source != "N":
            source = int(source)

        if dest != "N":
            dest = int(dest)

        sizes_d[source] += size
        sizes_d[dest] += size

    clusters_ids = []
    sizes = []
    for k, v in sorted(sizes_d.iteritems(), key=lambda x: x[0]):
        clusters_ids.append(str(k))
        sizes.append(v)

    flows = [{"source": clusters_ids.index(next(flow.iterkeys())), "dest": clusters_ids.index(next(flow.itervalues())),
              "value": size} for flow, size in zip(flows, flows_sizes) if 1. * size / sum(flows_sizes) >= threshold]

    if not labels:
        labels = clusters_ids
    else:
        tmp_labels = []
        threshold = 0.
        for id_ in clusters_ids:
            tmp_labels.append(labels[id_])

        labels = tmp_labels

    if colors:
        tmp_colors = []
        for id_ in clusters_ids:
            tmp_colors.append(colors[id_])

        colors = tmp_colors

    Chord(ax, 100, sizes, flows, labels, colors)

    return fig


# 7
def volume_scope_area(csv_processor):
    fig, ax = plt.subplots()
    ax.set_title("Scope area")
    ax.set_xlabel("Frame")
    ax.set_ylabel("Area")

    x = csv_processor.get_column_values("# frame")
    y = csv_processor.get_column_values("scope_area")

    ax.set_xlim((min(x), max(x)))
    ax.set_ylim((min(y), max(y)))

    ax.plot(x, y)

    return fig


# 7
def volume_scope_volume(csv_processor):
    fig, ax = plt.subplots()
    ax.set_title("Scope volume")
    ax.set_xlabel("Frame")
    ax.set_ylabel("Volume")

    x = csv_processor.get_column_values("# frame")
    y = csv_processor.get_column_values("scope_volume")

    ax.plot(x, y)

    ax.set_xlim((min(x), max(x)))
    ax.set_ylim((min(y), max(y)))

    return fig


# 7
def volume_object_area(csv_processor):
    fig, ax = plt.subplots()
    ax.set_title("Object area")
    ax.set_xlabel("Frame")
    ax.set_ylabel("Area")

    x = csv_processor.get_column_values("# frame")
    y = csv_processor.get_column_values("object_area")

    ax.plot(x, y)

    ax.set_xlim((min(x), max(x)))
    ax.set_ylim((min(y), max(y)))

    return fig


# 7
def volume_object_volume(csv_processor):
    fig, ax = plt.subplots()
    ax.set_title("Object volume")
    ax.set_xlabel("Frame")
    ax.set_ylabel("Area")

    x = csv_processor.get_column_values("# frame")
    y = csv_processor.get_column_values("object_volume")

    ax.plot(x, y)

    ax.set_xlim((min(x), max(x)))
    ax.set_ylim((min(y), max(y)))

    return fig


# 8
def cluster_area(file_processor, suffix=""):
    fig, ax = plt.subplots()

    ax.set_title("Clusters area" + suffix)
    ax.set_xlabel("Density")
    ax.set_ylabel("Area")

    column_names = file_processor.get_column_names("Clusters summary - areas" + suffix)
    clusters_id = file_processor.get_column_values("Clusters summary - areas" + suffix, "Cluster")

    x = [name for name in column_names if name.startswith("D")]
    x.sort(key=lambda x: int(x[1:]), reverse=True)

    # Values for each density
    column_values = [file_processor.get_column_values("Clusters summary - areas" + suffix, density) for density in x]

    # Transpose matrix to get values for each cluster
    y_set = zip(*column_values)

    for cluster_id, y in zip(clusters_id, y_set):
        ax.plot(x, y, label="Cluster {}".format(cluster_id))

    ax.set_xlim((0, len(x) - 1))
    ax.legend(bbox_to_anchor=(1.02, 1), loc="upper left")

    return fig


class StateFrame(tk.Frame):
    def __init__(self, parent, *args, **kw):
        tk.Frame.__init__(self, parent, *args, **kw)

        self.enabled = False

    def toggle(self):
        if self.enabled:
            self.enabled = False
            self.disable()
        else:
            self.enabled = True
            self.enable()

    def disable(self):
        self.enabled = False
        for child in self.winfo_children():
            child.configure(state=tk.DISABLED)

    def enable(self):
        self.enabled = True
        for child in self.winfo_children():
            child.configure(state=tk.NORMAL)


class Kraken(object):
    def __init__(self, parent):
        """
        Kraken App
        :param parent: Parent widget
        """
        self.parent = parent
        self.size = (600, 650)
        self.title = "Kraken"

        parent.title(self.title)
        parent.geometry("{}x{}".format(*self.size))

        logo = tk.PhotoImage(file=get_img("logo.gif"))

        logo_label = ttk.Label(self.parent, image=logo, padding=-2)
        logo_label.image = logo
        logo_label.pack(padx=20, pady=20)

        # Used to auto positioning depending on length of version string
        version = "ver. " + aquaduct.version_nice()
        ttk.Label(logo_label, text=version, background="white").place(relx=1, rely=1, x=-len(version) * 7, y=-20)

        self.main_frame = utils.VerticalScrolledFrame(self.parent)
        self.main_frame.pack(expand=1, fill="both")
        self.main_frame = self.main_frame.interior

        # General variables
        self.data_file = tk.StringVar()
        self.csv_file = tk.StringVar()
        self.dat_file = tk.StringVar()
        self.results_file = tk.StringVar()

        # Checkbuttons variables
        self.v1 = tk.BooleanVar()
        self.v2 = tk.BooleanVar()
        self.v3 = tk.BooleanVar()
        self.v4 = tk.BooleanVar()
        self.v7 = tk.BooleanVar()
        self.v8 = tk.BooleanVar()

        ###
        results_frame = tk.Frame(self.main_frame)
        results_frame.columnconfigure(0, weight=1)
        results_frame.columnconfigure(1, weight=1)
        results_frame.columnconfigure(2, weight=1)
        results_frame.pack(fill=tk.X, padx=100, pady=20)

        ttk.Label(results_frame, text="Output file: ").grid(sticky="e", row=2, column=0)
        ttk.Entry(results_frame, textvariable=self.results_file).grid(sticky="we", row=2, column=1)
        rload = ttk.Button(results_frame, text="Load file", style="File.TButton")
        rload.grid(sticky="w", row=2, column=2)
        rload.bind("<Button-1>", lambda e: self.load_file(self.results_file))

        # Container for plots which use txt
        container_data = tk.Frame(self.main_frame, bd=1, relief=tk.SUNKEN)
        container_data.columnconfigure(0, weight=1)
        container_data.columnconfigure(1, weight=1)
        container_data.columnconfigure(2, weight=1)
        container_data.pack(fill=tk.X, padx=100, pady=10)

        ttk.Label(container_data, text="Data file: ").grid(sticky="e", row=0, column=0)
        ttk.Entry(container_data, textvariable=self.data_file).grid(sticky="we", row=0, column=1)
        data_load = ttk.Button(container_data, text="Load file", style="File.TButton")
        data_load.grid(sticky="w", row=0, column=2)
        data_load.bind("<Button-1>", lambda e: self.load_file(self.data_file))

        # Container for plots which use csv
        container_csv = tk.Frame(self.main_frame, bd=1, relief=tk.SUNKEN)
        container_csv.columnconfigure(0, weight=1)
        container_csv.columnconfigure(1, weight=1)
        container_csv.columnconfigure(2, weight=1)
        container_csv.pack(fill=tk.X, padx=100, pady=10)

        ttk.Label(container_csv, text="CSV file: ").grid(sticky="e", row=0, column=0)
        ttk.Entry(container_csv, textvariable=self.csv_file).grid(sticky="we", row=0, column=1)
        csv_load = ttk.Button(container_csv, text="Load file", style="File.TButton")
        csv_load.grid(sticky="w", row=0, column=2)
        csv_load.bind("<Button-1>", lambda e: self.load_file(self.csv_file))

        ### 1
        self.all1 = tk.BooleanVar(value=1)
        self.molecules1 = tk.StringVar()

        option1_frame = tk.Frame(container_data, bd=1, relief=tk.GROOVE)
        option1_frame.columnconfigure(0, weight=1)
        option1_frame.columnconfigure(1, weight=1)
        option1_frame.grid(sticky="ew", row=1, column=0, columnspan=3, padx=10, pady=10)

        cb1 = ttk.Checkbutton(option1_frame, text="Clusters size", var=self.v1)
        cb1.pack(anchor="w")

        state1_frame = StateFrame(option1_frame)
        state1_frame.columnconfigure(0, weight=1)
        state1_frame.columnconfigure(1, weight=1)
        state1_frame.pack()

        cb1.configure(command=lambda: state1_frame.toggle())

        # Options
        ttk.Label(state1_frame, text="For all").grid(sticky="w", row=0, column=1)
        ttk.Checkbutton(state1_frame, var=self.all1).grid(sticky="e", row=0, column=0)

        ttk.Label(state1_frame, text="For molecules: ").grid(row=1, column=0)
        tk.Entry(state1_frame, textvariable=self.molecules1).grid(row=1, column=1)

        state1_frame.disable()

        ### 3
        self.all3 = tk.BooleanVar(value=1)
        self.molecules3 = tk.StringVar()

        option3_frame = tk.Frame(container_data, bd=1, relief=tk.GROOVE)
        option3_frame.columnconfigure(0, weight=1)
        option3_frame.columnconfigure(1, weight=1)
        option3_frame.grid(sticky="ew", row=2, column=0, columnspan=3, padx=10, pady=10)

        cb3 = ttk.Checkbutton(option3_frame, text="Molecule entry time distribution", var=self.v3)
        cb3.pack(anchor="w")

        state3_frame = StateFrame(option3_frame)
        state3_frame.columnconfigure(0, weight=1)
        state3_frame.columnconfigure(1, weight=1)
        state3_frame.pack()

        cb3.configure(command=lambda: state3_frame.toggle())

        # Options
        ttk.Label(state3_frame, text="For all").grid(sticky="w", row=0, column=1)
        ttk.Checkbutton(state3_frame, var=self.all3).grid(sticky="e", row=0, column=0)

        ttk.Label(state3_frame, text="For molecules: ").grid(row=1, column=0)
        tk.Entry(state3_frame, textvariable=self.molecules3).grid(row=1, column=1)

        state3_frame.disable()

        ### 4
        self.chord_threshold = tk.StringVar()
        self.clusters_info4 = tk.StringVar()
        option4_frame = tk.Frame(container_data, bd=1, relief=tk.GROOVE)
        option4_frame.columnconfigure(0, weight=1)
        option4_frame.columnconfigure(1, weight=1)
        option4_frame.grid(sticky="ew", row=3, column=0, columnspan=3, padx=10, pady=10)

        cb4 = ttk.Checkbutton(option4_frame, text="Intramolecular flows", var=self.v4)
        cb4.pack(anchor="w")

        state4_frame = StateFrame(option4_frame)
        state4_frame.columnconfigure(0, weight=1)
        state4_frame.columnconfigure(1, weight=1)
        state4_frame.pack()

        cb4.configure(command=lambda: state4_frame.toggle())

        ttk.Label(state4_frame, text="Clusters info: ").grid(row=0, column=0)

        tk.Entry(state4_frame, textvariable=self.clusters_info4).grid(row=0, column=1)
        clust_info_load = ttk.Button(state4_frame, command=lambda: self.load_file(self.clusters_info4),
                                     text="Load file", style="File.TButton")
        clust_info_load.grid(sticky="w", row=0, column=3)

        ttk.Label(state4_frame, text="Threshold: ").grid(row=1, column=0)
        tk.Entry(state4_frame, textvariable=self.chord_threshold).grid(row=1, column=1)

        state4_frame.disable()

        ### 8
        self.all8 = tk.StringVar(value=1)
        self.molecules8 = tk.StringVar()

        option8_frame = tk.Frame(container_data, bd=1, relief=tk.GROOVE)
        option8_frame.columnconfigure(0, weight=1)
        option8_frame.columnconfigure(1, weight=1)
        option8_frame.grid(sticky="ew", row=4, column=0, columnspan=3, padx=10, pady=10)

        cb8 = ttk.Checkbutton(option8_frame, text="Clusters area", var=self.v8)
        cb8.pack(anchor="w")

        state8_frame = StateFrame(option8_frame)
        state8_frame.columnconfigure(0, weight=1)
        state8_frame.columnconfigure(1, weight=1)
        state8_frame.pack()

        cb8.configure(command=lambda: state8_frame.toggle())

        # Options
        ttk.Label(state8_frame, text="For all").grid(sticky="w", row=0, column=1)
        ttk.Checkbutton(state8_frame, var=self.all8).grid(sticky="e", row=0, column=0)

        ttk.Label(state8_frame, text="For molecules: ").grid(row=1, column=0)
        tk.Entry(state8_frame, textvariable=self.molecules8).grid(row=1, column=1)

        state8_frame.disable()

        ### 2
        self.clusters_info2 = tk.StringVar()

        option2_frame = tk.Frame(container_csv, bd=1, relief=tk.GROOVE)
        option2_frame.grid(sticky="ew", row=1, column=0, columnspan=3, padx=10, pady=10)
        cb2 = ttk.Checkbutton(option2_frame, text="Relative cluster flows", var=self.v2)
        cb2.pack(anchor="w")

        state2_frame = StateFrame(option2_frame)
        state2_frame.columnconfigure(0, weight=1)
        state2_frame.columnconfigure(1, weight=1)
        state2_frame.pack()

        ttk.Label(state2_frame, text="Clusters info: ").grid(row=0, column=0)
        clust_info_load2 = ttk.Button(state2_frame, command=lambda: self.load_file(self.clusters_info2),
                                     text="Load file", style="File.TButton")
        clust_info_load2.grid(sticky="w", row=0, column=3)

        tk.Entry(state2_frame, textvariable=self.clusters_info2).grid(row=0, column=1)

        cb2.configure(command=lambda: state2_frame.toggle())

        state2_frame.disable()

        ### 7
        option7_frame = tk.Frame(container_csv, bd=1, relief=tk.GROOVE)
        option7_frame.grid(sticky="ew", row=2, column=0, columnspan=3, padx=10, pady=10)
        ttk.Checkbutton(option7_frame, text="Volume per frame", var=self.v7).pack(anchor="w")

        ###
        self.generate_button = ttk.Button(self.main_frame, text="Generate")
        self.generate_button.pack(pady=15)
        self.generate_button.bind("<Button-1>", lambda e: self.generate())

    def generate(self):
        if not True in [self.v1.get(), self.v2.get(), self.v3.get(), self.v4.get(), self.v7.get(), self.v8.get()]:
            tkMessageBox.showerror("Error", "No plot have been selected.")
            return

        if not self.results_file.get():
            tkMessageBox.showerror("Error", "You must specify results file.")
            return

        # Loading files
        traced_molecules = []
        if self.v1.get() or self.v3.get() or self.v4.get() or self.v8.get():
            if not self.data_file.get():
                tkMessageBox.showerror("Error", "Data file is not specified.")
                return

            try:
                f = FileDataProcessor(self.data_file.get())
            except Exception as e:
                tkMessageBox.showerror("Error", "Data file could not be opened.")
                print e

            # Fetching traced molecules names
            for line in f.file_lines:
                if line.startswith("Names of traced molecules:"):
                    traced_molecules = line.lstrip("Names of traced molecules:").split()

        if self.v2.get() or self.v7.get():
            if not self.csv_file.get():
                tkMessageBox.showerror("Error", "CSV file is not specified.")
                return

            try:
                c = CSVDataProcessor(self.csv_file.get())
            except Exception as e:
                tkMessageBox.showerror("Error", "CSV file could not be opened.")
                print e

        # Console log init
        log_window = tk.Toplevel(self.parent, width=100)
        log_console = tk.Text(log_window)
        log_console.pack(side=tk.LEFT, fill=tk.BOTH)

        log_scroll = tk.Scrollbar(log_window, command=log_console.yview)
        log_scroll.pack(side=tk.RIGHT, fill=tk.Y)
        log_console["yscrollcommand"] = log_scroll.set

        self.generate_button.config(state=tk.DISABLED)
        self.generate_button.unbind("<Button-1>")

        self.parent.update_idletasks()  # Force console log to show before generating plots

        def close_log_window():
            log_window.destroy()
            self.generate_button.config(state=tk.NORMAL)
            self.generate_button.bind("<Button-1>", lambda e: self.generate())

        log_window.protocol("WM_DELETE_WINDOW", close_log_window)

        # Log console settings
        log_console.tag_config("error", foreground="red")
        log_console.tag_config("success", foreground="green")

        def log(*args, **kwargs):
            log_console.config(state=tk.NORMAL)
            log_console.insert(*args, **kwargs)
            log_console.config(state=tk.DISABLED)

        # Plotting
        plots = []

        if self.v1.get():
            log(tk.END, "{}\nGenerating Clusters size\n".format("-" * 30))

            if self.all1.get():
                log(tk.END, "* All ")

                try:
                    plot = StringIO()
                    cluster_inlets(f).savefig(plot, format="png", bbox_inches="tight")
                    plots.append(plot)

                    log(tk.END, u"\u2714\n", "success")
                except DataException as e:
                    log(tk.END, u"\u2718\n", "error")
                    log(tk.END, "{}\n".format(e), "error")

            if self.molecules1.get():
                molecules = self.molecules1.get().replace(" ", "").upper().split(",")
                for molecule in molecules:
                    log(tk.END, "* {} ".format(molecule))

                    try:
                        plot = StringIO()
                        cluster_inlets(f, suffix=" of {}".format(molecule)).savefig(plot, format="png",
                                                                                    bbox_inches="tight")
                        plots.append(plot)

                        log(tk.END, u"\u2714\n", "success")
                    except DataException as e:
                        log(tk.END, u"\u2718\n", "error")
                        log(tk.END, "{}\n".format(e), "error")

            log(tk.END, "Done.\n", "success")

        if self.v2.get():
            log(tk.END, "{}\nGenerating Relative cluster flows\n".format("-" * 30))

            labels_file = {}
            colors_file = {}
            if self.clusters_info2.get():
                with open(self.clusters_info2.get(), "r") as file:
                    for line in file.readlines():
                        id_, name, color = line.rstrip().split("\t")
                        labels_file.update({id_: name})
                        colors_file.update({id_: color})

            # Finding all clusters
            clusters = []
            ids = []
            for column_name in c.column_names:
                match = re.match("amol_apaths_(\d+)_walk$", column_name)
                if match:
                    clusters.append(match.group(0))
                    ids.append(int(match.group(1)))

            ids, clusters = zip(*sorted(zip(ids, clusters)))

            labels = [labels_file.pop(str(i), i) for i in ids]
            cg = color_gen()
            colors = [colors_file.pop(str(i), next(cg)) for i in ids]

            plot = StringIO()
            relative_clusters_flows(c, clusters, labels, colors).savefig(plot, format="png", bbox_inches="tight")
            plots.append(plot)

            log(tk.END, "Done.\n", "success")

        if self.v3.get():
            log(tk.END, "{}\nGenerating Molecule entry time distribution\n".format("-" * 30))
            if self.all3.get():
                plot = StringIO()
                ligands_time(f).savefig(plot, format="png", bbox_inches="tight")
                plots.append(plot)

            if self.molecules3.get():
                molecules = self.molecules3.get().replace(" ", "").upper().split(",")
                for molecule in molecules:
                    log(tk.END, "* {} ".format(molecule))

                    plot = StringIO()
                    ligands_time(f, molecule).savefig(plot, format="png", bbox_inches="tight")
                    plots.append(plot)

                    log(tk.END, u"\u2714\n", "success")

            log(tk.END, "Done.\n", "success")

        if self.v4.get():
            log(tk.END, "{}\nGenerating Intramolecular flows\n".format("-" * 30))

            labels = {}
            colors = {}
            if self.clusters_info4.get():
                with open(self.clusters_info4.get(), "r") as file:
                    for line in file.readlines():
                        id_, name, color = line.rstrip().split("\t")
                        labels.update({id_: name})
                        colors.update({id_: color})

            threshold = float(self.chord_threshold.get()) if self.chord_threshold.get() else 0.0

            log(tk.END, "* Clusters sizes ")
            plot1 = StringIO()
            chord_diagram_sizes(f, labels, colors).savefig(plot1, format="png", dpi=2 ** 7, bbox_inches="tight")
            log(tk.END, u"\u2714\n", "success")

            log(tk.END, "* Clusters flows ")
            plot2 = StringIO()
            chord_diagram_flows(f, labels, colors, threshold).savefig(plot2, format="png", dpi=2 ** 7,
                                                                      bbox_inches="tight")
            log(tk.END, u"\u2714\n", "success")

            plots.extend([plot1, plot2])

            log(tk.END, "Done.\n", "success")

        if self.v7.get():
            log(tk.END, "{}\nGenerating Volumes\n".format("-" * 30))

            plot1 = StringIO()
            volume_scope_area(c).savefig(plot1, format="png", bbox_inches="tight")

            plot2 = StringIO()
            volume_scope_volume(c).savefig(plot2, format="png", bbox_inches="tight")

            plot3 = StringIO()
            volume_object_area(c).savefig(plot3, format="png", bbox_inches="tight")

            plot4 = StringIO()
            volume_object_volume(c).savefig(plot4, format="png", bbox_inches="tight")

            plots.extend([plot1, plot2, plot3, plot4])

            log(tk.END, "Done.\n", "success")

        if self.v8.get():
            log(tk.END, "{}\nGenerating Clusters areas\n".format("-" * 30))

            if self.all8.get():
                plot = StringIO()
                cluster_area(f).savefig(plot, format="png", bbox_inches="tight")
                plots.append(plot)

            if self.molecules8.get():
                molecules = self.molecules8.get().replace(" ", "").upper().split(
                    ",") if self.molecules8.get() else traced_molecules
                for molecule in molecules:
                    log(tk.END, "* {} ".format(molecule))

                    try:
                        plot = StringIO()
                        cluster_area(f, suffix=" of {}".format(molecule)).savefig(plot, format="png")
                        plots.append(plot)
                        log(tk.END, u"\u2714\n", "success")
                    except DataException as e:
                        log(tk.END, u"\u2718\n", "error")
                        log(tk.END, "{}\n".format(e), "error")

            log(tk.END, "Done.\n", "success")

        # Save plots to file
        logo_base64 = base64.b64encode(open(get_img("logo.gif"), "rb").read())
        html = """<div style="text-align:center; margin-bottom: 40px"><a href="http://www.tunnelinggroup.pl"><img src="data:image/gif;base64,{}"></a></div>""".format(
            logo_base64)
        for i, f in enumerate(plots):
            html += """<div style="text-align: center">
    <img src=\"data:image/png;base64,{}\">
</div>""".format(base64.encodestring(f.getvalue()))

        html += "<div style=\"text-align:center\"><h3>Document generated by <span style=\"color:#38b2fa\">Kraken</span>.</h3></div>"

        with open(self.results_file.get(), "w+") as f:
            f.write(html)
            log(tk.END, "Plots saved to {}.\n".format(self.results_file.get()), "success")

    def load_file(self, var):
        try:
            with askopenfile("r") as f:
                var.set(f.name)
        except AttributeError:  # In case of cancel selecting file
            pass


if __name__ == "__main__":
    root = tk.Tk()
    root.configure(background="white")
    root.resizable(1, 1)

    aq_icon = tk.PhotoImage(file=get_img("icon.gif"))
    root.tk.call('wm', 'iconphoto', root._w, aq_icon)

    s = ttk.Style()
    s.theme_use("clam")

    s.configure("TLabel", padding=5)
    s.configure("File.TButton", padding=0, font=("TkDefaultFont", 8))  # Loading file button

    app = Kraken(root)
    root.mainloop()
